﻿/*	File			DB_Arch_AppClient_for_SingleAppServer_DB1andDB2.cs
	Purpose			DB Client component for illustration of aspects of connectivity, coupling, architecture and transparency
	Author			Richard Anthony	(ar26@gre.ac.uk)
	Date			February 2015
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace DBArch_AppClient_for_SingleAppServer_DB1andDB2
{
    public partial class DB_Arch_AppClient_for_SingleAppServer_DB1andDB2 : Form
    {
        public const Int16 DB_TYPE1_PORT = 8002; // Decimal
        public const Int16 DB_TYPE2_PORT = 8003; // Decimal
        public const Int16 App_TYPE1_PORT = 8004; // Decimal

        public const Int32 DATABASE_SIZE = 3;
        public const Int16 MAX_CUST_NAME_LEN = 20;

        public enum MessageType {REQUEST_RECORD, SEND_RECORD, CLIENT_DISCONNECT};

        // Message structure (this will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Application_Message_PDU
        {
            public MessageType eMessageType;    // Offset in byte array 0
            public Int32 iCust_ID;              // Offset in byte array 4
            public bool bCustNameValid;         // Offset in byte array 8
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_CUST_NAME_LEN + 1))]
            public byte[] cCustName;	        // Offset in byte array 9 - 29
            public bool bCustBalanceValid;      // Offset in byte array 30
            // Single byte of padding exists in the byte arraay at this point, to align the following Int32 on a 4-byte boundary
            public Int32 iCustBalance;          // Offset in byte array 32
        };

        public const Int16 Application_Message_Size = 36;

        Socket m_ConnectSock_AppServer;
        IPEndPoint m_Connect_AppServer_IPEndPoint;

        bool m_bConnected_AppServer;
 
        System.Windows.Forms.Timer m_Timer_Application;

        // Data Row and Table structures (hard coded data serves as mock of real database)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct DataRow_struct 
        {
	        public int iCust_ID;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_CUST_NAME_LEN + 1))]
	        public byte[] cCustName;	
        };

        // DataTable_struct
	    DataRow_struct[] m_DataRowArray; // Array of data row structures

        public DB_Arch_AppClient_for_SingleAppServer_DB1andDB2()
        {
            InitializeComponent();
            m_DataRowArray = new DataRow_struct[DATABASE_SIZE];
            m_bConnected_AppServer = false;
            InitialiseControls();
            GetLocalHostAddressDetails();
        }

        void InitialiseControls()
        {	
            IP_1_textBox.Enabled = true;
            IP_2_textBox.Enabled = true;
            IP_3_textBox.Enabled = true;
            IP_4_textBox.Enabled = true;
            Port_textBox.Enabled = true;	
            Connect_button.Enabled = true;
            RequestData_button.Enabled = false;
        }

        void GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {   
                    // Initially display the local address (the first IPv4 address and the pre-defined port). 
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    IP_1_textBox.Text = bAddressBytes[0].ToString();
                    IP_2_textBox.Text = bAddressBytes[1].ToString();
                    IP_3_textBox.Text = bAddressBytes[2].ToString();
                    IP_4_textBox.Text = bAddressBytes[3].ToString();
                    Port_textBox.Text = App_TYPE1_PORT.ToString();
                    return;
                }
            }
            // No IPv4 address found - local host will not be able to connect to server using IPv4
            IP_1_textBox.Text = "No";
            IP_2_textBox.Text = "IP";
            IP_3_textBox.Text = "v4";
            IP_4_textBox.Text = "Addr";
            Connect_button.Enabled = false;
        }

        void GetConnectAddress_AppServer()
        {
            // Get the server address fron the user interface. Our own address is initially placed there to
            // reduce user interaction requirements (for example if both processes are on the same computer
            // no change to the address is required, or if both computers in the same Class C IP subnet
            // only the fourth address byte needs to be changed). The port number is also initialised on
            // the user interface to a default value, but must be read-in because the use may have changed it
            byte[] bAddressArray = new byte[4];
            bAddressArray[0] = Convert.ToByte(IP_1_textBox.Text);
            bAddressArray[1] = Convert.ToByte(IP_2_textBox.Text);
            bAddressArray[2] = Convert.ToByte(IP_3_textBox.Text);
            bAddressArray[3] = Convert.ToByte(IP_4_textBox.Text);
            Int16 iPort = Convert.ToInt16(Port_textBox.Text);
            try
            {
                IPAddress IP_Address = new IPAddress(bAddressArray);
                m_Connect_AppServer_IPEndPoint = new System.Net.IPEndPoint(IP_Address, iPort);
            }
            catch
            {
            }
        }

        void Start_Timer_Application()
        {
            m_Timer_Application = new System.Windows.Forms.Timer();
            m_Timer_Application.Interval = 100; // milliseconds
            m_Timer_Application.Tick += OnTimedEvent_Timer_Application;
            m_Timer_Application.Enabled = true;
        }

        void OnTimedEvent_Timer_Application(Object myObject, EventArgs myEventArgs)
        {
	        DoReceive();
        }

        void SendApplicationMessage(Application_Message_PDU Message, Socket Sock)
        {
            byte[] bBuffer = Serialize_Application_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = Sock.Send(bBuffer);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    if (false == Is_TCP_SocketConnected(Sock))
                    {   // Connection to server lost
                        CloseSocketAndExit();
                    }
                }
            }
         }

        void DoReceive()
        {		
            byte[] bArray = new byte[Application_Message_Size];
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;

            if (true == m_bConnected_AppServer)
            {
                iReceiveByteCount = 0;
                try
                {
                    iReceiveByteCount = m_ConnectSock_AppServer.Receive(bArray);
                }
                catch (SocketException ex)
                {
                    if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                    {
                        // IGNORE - The exception was 'WOULD BLOCK' 
                    }
                    else
                    {
                        // Additional diagnostics can be placed here if necessary
                    }
                }
                if (0 < iReceiveByteCount)
                {
                    Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
                    switch (Message.eMessageType)
                    {
                        case MessageType.SEND_RECORD:
                            DisplayData(Message);
                            break;
                    }
                }
            }
        }

        Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
            Application_Message_PDU result_Message = new Application_Message_PDU();
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0/*FieldOffset*/);
            result_Message.iCust_ID = ByteArrayToInt32(bRecdByteArray, 4/*FieldOffset*/);
            result_Message.bCustNameValid = true;
            if (0 == bRecdByteArray[8])
            {
                result_Message.bCustNameValid = false;
            }
            result_Message.cCustName = new byte[MAX_CUST_NAME_LEN+1];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 9/*source offset*/, result_Message.cCustName/*dest*/, 0/*cCustName offset*/, MAX_CUST_NAME_LEN/*count*/);
            result_Message.bCustBalanceValid = true;
            if (0 == bRecdByteArray[30])
            {
                result_Message.bCustBalanceValid = false;
            }
            result_Message.iCustBalance = ByteArrayToInt32(bRecdByteArray, 32/*FieldOffset*/);
            return result_Message;
        }

        private byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
        {   // Convert an Application_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[Application_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 0/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCust_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 4/*bByteArray offset*/, 4/*count*/);
            bByteArray[8] = 0; // false (bCustNameValid)
            if(true == Message.bCustNameValid)
            {
                bByteArray[8] = 1; // true (bCustNameValid)
            }
            if (null != Message.cCustName)
            {
                Buffer.BlockCopy(Message.cCustName /*src*/, 0/*source offset*/, bByteArray/*dest*/, 9/*cCustName offset*/, Message.cCustName.Length/*count*/);
            }
            bByteArray[30] = 0; // false (bCustBalanceValid)
            if (true == Message.bCustBalanceValid)
            {
                bByteArray[30] = 1; // true (bCustBalanceValid)
            }
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCustBalance);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 32/*bByteArray offset*/, 4/*count*/);
            return bByteArray;
        }

        static bool Is_TCP_SocketConnected(Socket sSock)
        {
            if (sSock.ProtocolType != ProtocolType.Tcp)
            {   // This test is necessary because the Connected property may return true for a UDP
                // socket (which is meaningless as it is alsways 'connectionless')
                return false;
            }
            // Note, the Connected property alone is not reliable. It indicates the connected state of the socket
            // at the time of executing the most recent Send or Receive operation
            return !((sSock.Poll(1000, SelectMode.SelectRead) && (sSock.Available == 0)) || !sSock.Connected);
        }

        private void Done_Button_Click(object sender, EventArgs e)
        {
            CloseSocketAndExit();
        }

        void CloseSocketAndExit()
        {
            CloseConnectionAndSocket();
            System.Environment.Exit(-1);
        }

        void CloseConnections()
        {
            try
            {
                m_ConnectSock_AppServer.Shutdown(SocketShutdown.Both);	// Close the connection (in both directions)
            }
            catch
            {   // Generic approach to dealing with situations such as socket not connected
            }
        }

        void CloseConnectionAndSocket()
        {
            if(null != m_Timer_Application)
            {
                m_Timer_Application.Enabled = false;
            }
            CloseConnections();
            try
            {
                m_ConnectSock_AppServer.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if (sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }

        private Int32 GetCustIDFromControl()
        {
            Int32 iCustID = 0;
            if ("" == CustID_textBox.Text)
            {
                MessageBox.Show("Customer ID must be between 101 and 103", "Invalid Customer ID");
                return -1;
            }
            try
            {
                iCustID = Convert.ToInt32(CustID_textBox.Text);
                if (iCustID < 101 || iCustID > 103)
                {
                    MessageBox.Show("Customer ID must be between 101 and 103", "Invalid Customer ID");
                    return -1;
                }
            }
            catch
            {
                MessageBox.Show("Please enter a valid customer ID", "Invalid Customer ID");
                return -1;
            }
            return iCustID;
        }    

        private void RequestData_button_Click(object sender, EventArgs e)
        {

            Int32 iCustID = GetCustIDFromControl();
            if (-1 == iCustID)
            {
                return;
            }
            CustName_textBox.Text = "";   // Initially clear the data control
            Application_Message_PDU Message = new Application_Message_PDU();
            Message.eMessageType = MessageType.REQUEST_RECORD;
            Message.iCust_ID = iCustID;
            Message.bCustNameValid = false;
            if (true == m_bConnected_AppServer)
	        {
                SendApplicationMessage(Message, m_ConnectSock_AppServer);  // Send REQUEST_RECORD message to App server
	        }
        }

        void CreateConnectSocket_AndSetBlocking()
        {
            try
            {
                m_ConnectSock_AppServer = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ConnectSock_AppServer.Blocking = true; // Set blocking mode for socket during connection set-up phase
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "DB APP CLIENT");
                CloseSocketAndExit();
            }
        }

        private void Connect_button_Click(object sender, EventArgs e)
        {	
            CreateConnectSocket_AndSetBlocking();
            GetConnectAddress_AppServer();
            try
            {
                m_ConnectSock_AppServer.Connect(m_Connect_AppServer_IPEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Connect failed !\n\nIs server address correct ?\nIs port correct ?\nIs server running ?", "DB APP CLIENT");
                return;
            }

            m_bConnected_AppServer = true;
            RequestData_button.Enabled = true;

            IP_1_textBox.Enabled = false;
            IP_2_textBox.Enabled = false;
            IP_3_textBox.Enabled = false;
            IP_4_textBox.Enabled = false;
            Port_textBox.Enabled = false;
            Connect_button.Enabled = false;

            // Set socket non-blocking (now that connection established) for receiving
            try
            {
                m_ConnectSock_AppServer.Blocking = false;
            }
            catch(SocketException)
		    {
                MessageBox.Show("Could not set non-Blocking behaviour for socket","DB APP CLIENT");
                CloseSocketAndExit();
            }
            Start_Timer_Application();
        }

        void DisplayData(Application_Message_PDU Message)
        {
	        if(true == Message.bCustNameValid)		// Guard needed because each server returns only a partial row of data at the application level
	        {
		        string sStr = "";
                ByteArrayTo_ASCII_String(Message.cCustName, ref sStr);
                CustName_textBox.Text = sStr;
	        }
	        if(true == Message.bCustBalanceValid)	// Guard needed because each server returns only a partial row of data at the application level
	        {
		        CustBalance_textBox.Text = Message.iCustBalance.ToString();
	        }
        }
    }
}
