﻿/*	File			DB_Arch_AppServer_for_DB2.cs
	Purpose			Application SERVER component for illustration of aspects of connectivity, coupling, architecture and transparency
	Author			Richard Anthony	(ar26@gre.ac.uk)
	Date			February 2015
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace DBArch_AppServer_for_DB2
{
    public partial class DB_Arch_AppServer_for_DB2 : Form
    {
        public const Int16 DB_TYPE1_PORT = 8002; // Decimal
        public const Int16 DB_TYPE2_PORT = 8003; // Decimal
        public const Int16 App_TYPE1_PORT = 8004; // Decimal
        public const Int16 App_TYPE2_PORT = 8005; // Decimal

        public const Int32 DATABASE_SIZE = 3;
        public const Int16 MAX_CUST_NAME_LEN = 20;

        public enum MessageType {REQUEST_RECORD, SEND_RECORD, CLIENT_DISCONNECT};

        // Message structure (this will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Application_Message_PDU
        {
            public MessageType eMessageType;    // Offset in byte array 0
            public Int32 iCust_ID;              // Offset in byte array 4
            public bool bCustNameValid;         // Offset in byte array 8
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_CUST_NAME_LEN + 1))]
            public byte[] cCustName;	        // Offset in byte array 9 - 29
            public bool bCustBalanceValid;      // Offset in byte array 30
            // Single byte of padding exists in the byte arraay at this point, to align the following Int32 on a 4-byte boundary
            public Int32 iCustBalance;          // Offset in byte array 32
        };

        public const Int16 Application_Message_Size = 36;

        Socket m_ListenSock_AppClient;
        Socket m_ConnectSock_AppClient;
        Socket m_ConnectSock_DB_Server;
        IPEndPoint m_Local_IPEndPoint;
        IPEndPoint m_ConnectDB_IPEndPoint;

        bool m_bConnected_DB_Server;
        bool m_bConnected_App_Client;

        System.Windows.Forms.Timer m_Timer_Application;

        public DB_Arch_AppServer_for_DB2()
        {
            InitializeComponent();
            m_bConnected_DB_Server = false;
            m_bConnected_App_Client = false;
            GetLocalHostAddressDetails();
	        EnableServer();
        }

        void GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {   // Create the local IPEndPoint using the first IPv4 address and the pre-defined port
                    m_Local_IPEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], App_TYPE2_PORT);
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    string sStr = string.Format("Server Address (application):{0}.{1}.{2}.{3}:{4}", 
                        bAddressBytes[0].ToString(), bAddressBytes[1].ToString(), bAddressBytes[2].ToString(),
                        bAddressBytes[3].ToString(), App_TYPE2_PORT);
                    WriteStatusLine(sStr);
                    LocalAddress_textBox.Text = ipHostAddress[iAddrIndex].ToString();

                    m_ConnectDB_IPEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], DB_TYPE2_PORT);
                    AppServerPort_textBox.Text = App_TYPE2_PORT.ToString();
                    DBServerPort_textBox.Text = DB_TYPE2_PORT.ToString();
                    return;
                }
            }
	        MessageBox.Show("Failed to get local address!","DB_Arch_AppServer_for_DB2");
            CloseSocketAndExit();
        }
  
        void Accept()
        {
            try
            {
                m_ConnectSock_AppClient = m_ListenSock_AppClient.Accept();
                // Set the newly-created connection socket to Non-Blocking
                m_ConnectSock_AppClient.Blocking = false;
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    MessageBox.Show("Accept failed","DB_Arch_AppServer_for_DB2");
                }
                return;
            }
            m_bConnected_App_Client = true;
            string sStr = string.Format("Client {0} Connected", m_ConnectSock_AppClient.RemoteEndPoint.ToString());
            WriteStatusLine(sStr);
        }

        void EnableServer()  // Invoked on startup
        {
            try
            {   // Create Listen Socket and set it NonBlocking
                m_ListenSock_AppClient = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ListenSock_AppClient.Blocking = false;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "DB_Arch_AppServer_for_DB2");
                CloseSocketAndExit();
            }

            try
            {
                m_ListenSock_AppClient.Bind(m_Local_IPEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Bind failed !\nPort is probably in use","DB_Arch_AppServer_for_DB2");
                CloseSocketAndExit();
            }

	        // Listen - for TCP connection requests from clients
            try
            {
                m_ListenSock_AppClient.Listen(5);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Listen failed","DB_Arch_AppServer_for_DB2");
                CloseSocketAndExit();
            }
            Start_Timer_Application();   // As all activity is non-blocking
        }

        void Start_Timer_Application()
        {
            m_Timer_Application = new System.Windows.Forms.Timer();
            m_Timer_Application.Interval = 100; // milliseconds
            m_Timer_Application.Tick += OnTimedEvent_Timer_Application;
            m_Timer_Application.Enabled = true;
        }

        void OnTimedEvent_Timer_Application(Object myObject, EventArgs myEventArgs)
        {
	        if(false == m_bConnected_App_Client)
	        {
		        Accept(); // This only applies to the App Client connection
	        }
	
	        DoReceive(); // This applies to both connections
        }

        void Send_To_AppClient(Application_Message_PDU Message)
        {
            byte[] bBuffer = Serialize_Application_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_ConnectSock_AppClient.Send(bBuffer);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    if (false == Is_TCP_SocketConnected(m_ConnectSock_AppClient))
                    {   // Connection to server lost
                        WriteStatusLine("Connection lost");
                        CloseSocketAndExit();
                    }
                }
            }
        }

        void Send_To_DB_Server(Application_Message_PDU Message)
        {
            byte[] bBuffer = Serialize_Application_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_ConnectSock_DB_Server.Send(bBuffer);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    if (false == Is_TCP_SocketConnected(m_ConnectSock_DB_Server))
                    {   // Connection to server lost
                        WriteStatusLine("Connection lost");
                        CloseSocketAndExit();
                    }
                }
            }
         }
 				
        void Forward_DataRow_Message_To_AppClient(Application_Message_PDU Message_DB_In)
        {
	        Application_Message_PDU Message_App_Out = new Application_Message_PDU();
            Message_App_Out.cCustName = new byte[MAX_CUST_NAME_LEN + 1];
	        switch(Message_DB_In.eMessageType)
	        {
		        case MessageType.SEND_RECORD:
			        Message_App_Out.eMessageType = MessageType.SEND_RECORD;
			        Message_App_Out.iCust_ID = Message_DB_In.iCust_ID;
			        Message_App_Out.bCustBalanceValid = Message_DB_In.bCustBalanceValid;
			        Message_App_Out.bCustNameValid = Message_DB_In.bCustNameValid;
		
			        if(true == Message_DB_In.bCustBalanceValid)
			        {	
				        Message_App_Out.iCustBalance = Message_DB_In.iCustBalance;
			        }
			
			        if(true == Message_DB_In.bCustNameValid)
			        {	
                        Buffer.BlockCopy(Message_DB_In.cCustName/*src*/, 0/*offset*/, 
                            Message_App_Out.cCustName/*dest*/, 0/*offset*/, Message_DB_In.cCustName.Length/*count*/);
			        }
						
			        Send_To_AppClient(Message_App_Out);		
			        string sStr = string.Format("Data Record {0} Forwarded to App Client", Message_App_Out.iCust_ID);
			        WriteStatusLine(sStr);
			        break;
	        }
        }

        void Forward_Request_Message_To_DB_Server(Application_Message_PDU Message_App_In)
        {
	        // This method hides details of the spceific database request format from the client
	        // The client makes its request in an application-defined manner, independent of the format used by the database
	        Application_Message_PDU Message_DB_Out = new Application_Message_PDU();
            Message_DB_Out.cCustName = new byte[MAX_CUST_NAME_LEN + 1];
	        switch(Message_App_In.eMessageType)
	        {
		        case MessageType.REQUEST_RECORD:
			        Message_DB_Out.eMessageType = MessageType.REQUEST_RECORD;
			        Message_DB_Out.iCust_ID = Message_App_In.iCust_ID;
			        Send_To_DB_Server(Message_DB_Out);		
			        string sStr = string.Format("Data Record {0} request forwarded to DB", Message_DB_Out.iCust_ID);
			        WriteStatusLine(sStr);			
			        break;
	        }
        }

        void DoReceive()
        {		
            byte[] bArray = new byte[Application_Message_Size];
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;

	        // Check for Application messages (from DB server)
	        if(true == m_bConnected_DB_Server)
	        {
                iReceiveByteCount = 0;
                try
                {
                    iReceiveByteCount = m_ConnectSock_DB_Server.Receive(bArray);
                }
                catch (SocketException ex)
                {
                    if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                    {
                        // IGNORE - The exception was 'WOULD BLOCK' 
                    }
                    else
                    {
                        // Additional diagnostics can be placed here if necessary
                    }
                }
			    if(0 < iReceiveByteCount)
			    {
                    Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
			        switch(Message.eMessageType)
			        {
				    case MessageType.SEND_RECORD:
					    string sStr = string.Format("Data Record {0} received", Message.iCust_ID);
					    WriteStatusLine(sStr);
	                    if(true == m_bConnected_App_Client)
	                    {
					        Forward_DataRow_Message_To_AppClient(Message);
                        }
					    break;
			        }
		        }
	        }
        
 	        // Check for Application messages (from App client)
	        if(true == m_bConnected_App_Client)
	        {
                iReceiveByteCount = 0;
                try
                {
                    iReceiveByteCount = m_ConnectSock_AppClient.Receive(bArray);
                }
                catch (SocketException ex)
                {
                    if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                    {
                        // IGNORE - The exception was 'WOULD BLOCK' 
                    }
                    else
                    {
                        // Additional diagnostics can be placed here if necessary
                    }
                    return;
                }
			    if(0 < iReceiveByteCount)
			    {
                    Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
			        switch(Message.eMessageType)
			        {
				       case MessageType.REQUEST_RECORD:
					        string sStr = string.Format("Data record {0} requested", Message.iCust_ID);
					        WriteStatusLine(sStr);
				            if(true == m_bConnected_DB_Server)
				            {
					            Forward_Request_Message_To_DB_Server(Message);
				            }
					        break;		

				        case MessageType.CLIENT_DISCONNECT:
					        m_bConnected_App_Client = false;
					        CloseConnection();
					        WriteStatusLine("Client disconnected");
					        break;
			        }
		        }
	        }
        }

        private void Connect_button_Click(object sender, EventArgs e)
        {
	        // Create socket
            try
            {   // Create Socket for connecting to the DB server and set it Blocking
                m_ConnectSock_DB_Server = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ConnectSock_DB_Server.Blocking = true;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "DB_Arch_AppServer_for_DB2");
                CloseSocketAndExit();
            }

	        // Connect
            try
            {
                m_ConnectSock_DB_Server.Connect(m_ConnectDB_IPEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Connect failed !\n\nIs server address correct ?\nIs port correct ?\nIs server running ?", "DB_Arch_AppServer_for_DB2");
                return;
            }

            // Set non-blocking (now that connection established) for receiving
            m_ConnectSock_DB_Server.Blocking = false;
            
            WriteStatusLine("Connected to DB server");
	        m_bConnected_DB_Server = true;
            Connect_button.Enabled = false;
        }

        private void ClearLog_button_Click(object sender, EventArgs e)
        {
            EventLog_listBox.Items.Clear();
        }

        Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
            Application_Message_PDU result_Message = new Application_Message_PDU();
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0/*FieldOffset*/);
            result_Message.iCust_ID = ByteArrayToInt32(bRecdByteArray, 4/*FieldOffset*/);
            result_Message.bCustNameValid = true;
            if (0 == bRecdByteArray[8])
            {
                result_Message.bCustNameValid = false;
            }
            result_Message.cCustName = new byte[MAX_CUST_NAME_LEN+1];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 9/*source offset*/, result_Message.cCustName/*dest*/, 0/*cCustName offset*/, MAX_CUST_NAME_LEN/*count*/);
            result_Message.bCustBalanceValid = true;
            if (0 == bRecdByteArray[30])
            {
                result_Message.bCustBalanceValid = false;
            }
            result_Message.iCustBalance = ByteArrayToInt32(bRecdByteArray, 32/*FieldOffset*/);
            return result_Message;
        }

        private byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
        {   // Convert an Application_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[Application_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 0/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCust_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 4/*bByteArray offset*/, 4/*count*/);
            bByteArray[8] = 0; // false (bCustNameValid)
            if(true == Message.bCustNameValid)
            {
                bByteArray[8] = 1; // true (bCustNameValid)
            }
            if (null != Message.cCustName)
            {
                Buffer.BlockCopy(Message.cCustName /*src*/, 0/*source offset*/, bByteArray/*dest*/, 9/*cCustName offset*/, Message.cCustName.Length/*count*/);
            }
            bByteArray[30] = 0; // false (bCustBalanceValid)
            if (true == Message.bCustBalanceValid)
            {
                bByteArray[30] = 1; // true (bCustBalanceValid)
            }
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCustBalance);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 32/*bByteArray offset*/, 4/*count*/);
            return bByteArray;
        }

        void WriteStatusLine(string sLine)
        {
            EventLog_listBox.Items.Add(sLine);
        }

        static bool Is_TCP_SocketConnected(Socket sSock)
        {
            if (sSock.ProtocolType != ProtocolType.Tcp)
            {   // This test is necessary because the Connected property may return true for a UDP
                // socket (which is meaningless as it is alsways 'connectionless')
                return false;
            }
            // Note, the Connected property alone is not reliable. It indicates the connected state of the socket
            // at the time of executing the most recent Send or Receive operation
            return !((sSock.Poll(1000, SelectMode.SelectRead) && (sSock.Available == 0)) || !sSock.Connected);
        }

        private void Done_Button_Click(object sender, EventArgs e)
        {
            if (true == m_bConnected_DB_Server)
            {
                Application_Message_PDU Message_DB_Out = new Application_Message_PDU();
                Message_DB_Out.eMessageType = MessageType.CLIENT_DISCONNECT;// The App Server is a 'client' as seen by the DB Server
                Message_DB_Out.cCustName = new byte[MAX_CUST_NAME_LEN + 1];
                Send_To_DB_Server(Message_DB_Out);
            }
            CloseSocketAndExit();
        }

        void CloseSocketAndExit()
        {
            CloseConnectionAndSocket();
            System.Environment.Exit(-1);
        }

        void CloseConnection()
        {
            try
            {
                m_ConnectSock_AppClient.Shutdown(SocketShutdown.Both);	// Close the connection (in both directions)
            }
            catch
            {   // Generic approach to dealing with situations such as socket not connected
            }
            try
            {
                m_ConnectSock_DB_Server.Shutdown(SocketShutdown.Both);	// Close the connection (in both directions)
            }
            catch
            {   // Generic approach to dealing with situations such as socket not connected
            }
        }

        void CloseConnectionAndSocket()
        {
            if(null != m_Timer_Application)
            {
                m_Timer_Application.Enabled = false;
            }
            CloseConnection();
            try
            {
                m_ConnectSock_AppClient.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
            try
            {
                m_ConnectSock_DB_Server.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if (sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }
    }
}
