﻿/*	File			DB_Arch_DBServer_DB1_Direct.cs
	Purpose			DB SERVER component for illustration of aspects of connectivity, coupling, architecture and transparency
	Author			Richard Anthony	(ar26@gre.ac.uk)
	Date			February 2015
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace DBArch_DBServer_DB1_Direct
{
    public partial class DB_Arch_DBServer_DB1_Direct : Form
    {
        public const Int16 DB_TYPE1_PORT = 8002; // Decimal
        public const Int16 DB_TYPE2_PORT = 8003; // Decimal

        public const Int32 DATABASE_SIZE = 3;
        public const Int16 MAX_CUST_NAME_LEN = 20;

        public enum MessageType {REQUEST_RECORD, SEND_RECORD, CLIENT_DISCONNECT};

        // Message structure (this will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Application_Message_PDU
        {
            public MessageType eMessageType;    // Offset in byte array 0
            public Int32 iCust_ID;              // Offset in byte array 4
            public bool bCustNameValid;         // Offset in byte array 8
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_CUST_NAME_LEN + 1))]
            public byte[] cCustName;	        // Offset in byte array 9 - 29
            public bool bCustBalanceValid;      // Offset in byte array 30
            // Single byte of padding exists in the byte arraay at this point, to align the following Int32 on a 4-byte boundary
            public Int32 iCustBalance;          // Offset in byte array 32
        };

        public const Int16 Application_Message_Size = 36;

	    Socket m_ListenSock;  // Listen for client connections
	    Socket m_ConnectSock; // Connection to Client
 	    IPEndPoint m_Local_IPEndPoint;
	    IPEndPoint m_Connect_IPEndPoint;

	    bool m_bConnected;

        System.Windows.Forms.Timer m_Timer_Application;

        // Data Row and Table structures (hard coded data serves as mock of real database)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct DataRow_struct 
        {
	        public int iCust_ID;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_CUST_NAME_LEN + 1))]
	        public byte[] cCustName;	
        };

        // DataTable_struct
	    DataRow_struct[] m_DataRowArray; // Array of data row structures

        public DB_Arch_DBServer_DB1_Direct()
        {
            InitializeComponent();
            m_DataRowArray = new DataRow_struct[DATABASE_SIZE];
            GetLocalHostAddressDetails();
          
	        m_bConnected = false;

	        InitialiseDatabase();
	        DisplayDataTable();
	        EnableServer();
        }

        void InitialiseDatabase()	// Populate DataTable (hardcoded data)
        {
            m_DataRowArray[0].iCust_ID = 101;
            m_DataRowArray[0].cCustName = StringTo_ASCII_ByteArray("Fred");
            m_DataRowArray[1].iCust_ID = 102;
            m_DataRowArray[1].cCustName = StringTo_ASCII_ByteArray("Jane");
            m_DataRowArray[2].iCust_ID = 103;
            m_DataRowArray[2].cCustName = StringTo_ASCII_ByteArray("Rita");
        }

        void DisplayDataTable()
        {	
	        CustID_listBox.Items.Clear();
	        CustName_listBox.Items.Clear();

	        for(int iIndex = 0; iIndex < DATABASE_SIZE; iIndex ++)
	        {
		        CustID_listBox.Items.Add(m_DataRowArray[iIndex].iCust_ID.ToString());
                string sStr = "";
                ByteArrayTo_ASCII_String(m_DataRowArray[iIndex].cCustName, ref sStr);
                CustName_listBox.Items.Add(sStr);
            }
        }

        void GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {   // Create the local IPEndPoint using the first IPv4 address and the pre-defined port
                    m_Local_IPEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], DB_TYPE1_PORT);
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    string sStr = string.Format("Server Address (application):{0}.{1}.{2}.{3}:{4}", 
                        bAddressBytes[0].ToString(), bAddressBytes[1].ToString(), bAddressBytes[2].ToString(),
                        bAddressBytes[3].ToString(), DB_TYPE1_PORT);
                    WriteStatusLine(sStr);
                 
                    return;
                }
            }
	        MessageBox.Show("Failed to get local address!","CDB_Arch_DBServer");
            CloseSocketAndExit();
        }
        
       void Accept()
        {
            try
            {
                m_ConnectSock = m_ListenSock.Accept();
                // Set the newly-created connection socket to Non-Blocking
                m_ConnectSock.Blocking = false;
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    MessageBox.Show("Accept failed","CDB_Arch_DBServer");
                }
                return;
            }
            m_bConnected = true;
            string sStr = string.Format("Client {0} Connected", m_ConnectSock.RemoteEndPoint.ToString());
            WriteStatusLine(sStr);
        }

        void EnableServer()  // Invoked on startup
        {
            try
            {   // Create Listen Socket and set it NonBlocking
                m_ListenSock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ListenSock.Blocking = false;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "CDB_Arch_DBServer");
                CloseSocketAndExit();
            }

            try
            {
                m_ListenSock.Bind(m_Local_IPEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Bind failed !\nPort is probably in use","CDB_Arch_DBServer");
                CloseSocketAndExit();
            }

	        // Listen - for TCP connection requests from clients
            try
            {
                m_ListenSock.Listen(5);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Listen failed","CDB_Arch_DBServer");
                CloseSocketAndExit();
            }
            Start_Timer_Application();   // As all activity is non-blocking
        }
 
        void Start_Timer_Application()
        {
            m_Timer_Application = new System.Windows.Forms.Timer();
            m_Timer_Application.Interval = 100; // milliseconds
            m_Timer_Application.Tick += OnTimedEvent_Timer_Application;
            m_Timer_Application.Enabled = true;
        }

        void OnTimedEvent_Timer_Application(Object myObject, EventArgs myEventArgs)
        {
	        if(true == m_bConnected)
	        {
		        DoReceive();
	        }
	        else
	        {
		        Accept();
	        }
        }

        void SendApplicationMessage(Application_Message_PDU Message)
        {
            byte[] bBuffer = Serialize_Application_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_ConnectSock.Send(bBuffer);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    if (false == Is_TCP_SocketConnected(m_ConnectSock))
                    {   // Connection to server lost
                        WriteStatusLine("Connection lost");
                        CloseSocketAndExit();
                    }
                }
            }
         }

        void Send_DataRow_Message(int iCustID)
        {
	        Application_Message_PDU Message = new Application_Message_PDU();
	        Message.eMessageType = MessageType.SEND_RECORD;
	        Message.iCust_ID = iCustID;
	        Message.cCustName = StringTo_ASCII_ByteArray("------No Match------");

	        for(int iIndex = 0; iIndex < DATABASE_SIZE; iIndex++)
	        {
		        if(iCustID == m_DataRowArray[iIndex].iCust_ID)
		        {
                    Buffer.BlockCopy(m_DataRowArray[iIndex].cCustName/*src*/, 0 /*offset*/, Message.cCustName/*dest*/, 0/*offset*/, m_DataRowArray[iIndex].cCustName.Length/*count*/);
                }
	        }

	        Message.bCustNameValid = true;
	        Message.bCustBalanceValid = false;
	        SendApplicationMessage(Message);
            string sCust = "";
            ByteArrayTo_ASCII_String(Message.cCustName, ref sCust);
            string sStr = string.Format("Data Record {0} {1} sent", Message.iCust_ID, sCust);
	        WriteStatusLine(sStr);
        }

        void DoReceive()
        {		
	        // Check for Application messages (from Client)
            byte[] bArray = new byte[Application_Message_Size];
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;

	        if(true == m_bConnected)
	        {
                iReceiveByteCount = 0;
                try
                {
                    iReceiveByteCount = m_ConnectSock.Receive(bArray);
                }
                catch (SocketException ex)
                {
                    if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                    {
                        // IGNORE - The exception was 'WOULD BLOCK' 
                    }
                    else
                    {
                        // Additional diagnostics can be placed here if necessary
                    }
                    return;
                }
			    if(0 < iReceiveByteCount)
			    {
                    Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
			        string sStr;
			        switch(Message.eMessageType)
			        {
				       case MessageType.REQUEST_RECORD:
					        sStr = string.Format("Data record {0} requested", Message.iCust_ID);
					        WriteStatusLine(sStr);
					        Send_DataRow_Message(Message.iCust_ID);
					        break;		

				        case MessageType.CLIENT_DISCONNECT:
					        m_bConnected = false;
					        CloseConnection();
					        WriteStatusLine("Client disconnected");
					        break;
			        }
		        }
	        }
        }

        private void ClearLog_button_Click(object sender, EventArgs e)
        {
            EventLog_listBox.Items.Clear();
        }

        Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
            Application_Message_PDU result_Message = new Application_Message_PDU();
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0/*FieldOffset*/);
            result_Message.iCust_ID = ByteArrayToInt32(bRecdByteArray, 4/*FieldOffset*/);
            result_Message.bCustNameValid = true;
            if (0 == bRecdByteArray[8])
            {
                result_Message.bCustNameValid = false;
            }
            result_Message.cCustName = new byte[MAX_CUST_NAME_LEN+1];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 9/*source offset*/, result_Message.cCustName/*dest*/, 0/*cCustName offset*/, MAX_CUST_NAME_LEN/*count*/);
            result_Message.bCustBalanceValid = true;
            if (0 == bRecdByteArray[30])
            {
                result_Message.bCustBalanceValid = false;
            }
            result_Message.iCustBalance = ByteArrayToInt32(bRecdByteArray, 32/*FieldOffset*/);
            return result_Message;
        }

        private byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
        {   // Convert an Application_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[Application_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 0/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCust_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 4/*bByteArray offset*/, 4/*count*/);
            bByteArray[8] = 0; // false (bCustNameValid)
            if(true == Message.bCustNameValid)
            {
                bByteArray[8] = 1; // true (bCustNameValid)
            }
            if (null != Message.cCustName)
            {
                Buffer.BlockCopy(Message.cCustName /*src*/, 0/*source offset*/, bByteArray/*dest*/, 9/*cCustName offset*/, Message.cCustName.Length/*count*/);
            }
            bByteArray[30] = 0; // false (bCustBalanceValid)
            if (true == Message.bCustBalanceValid)
            {
                bByteArray[30] = 1; // true (bCustBalanceValid)
            }
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCustBalance);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 32/*bByteArray offset*/, 4/*count*/);
            return bByteArray;
        }

        void WriteStatusLine(string sLine)
        {
            EventLog_listBox.Items.Add(sLine);
        }

        static bool Is_TCP_SocketConnected(Socket sSock)
        {
            if (sSock.ProtocolType != ProtocolType.Tcp)
            {   // This test is necessary because the Connected property may return true for a UDP
                // socket (which is meaningless as it is alsways 'connectionless')
                return false;
            }
            // Note, the Connected property alone is not reliable. It indicates the connected state of the socket
            // at the time of executing the most recent Send or Receive operation
            return !((sSock.Poll(1000, SelectMode.SelectRead) && (sSock.Available == 0)) || !sSock.Connected);
        }

        private void Done_Button_Click(object sender, EventArgs e)
        {
            CloseSocketAndExit();
        }

        void CloseSocketAndExit()
        {
            CloseConnectionAndSocket();
            System.Environment.Exit(-1);
        }

        void CloseConnection()
        {
            try
            {
                m_ConnectSock.Shutdown(SocketShutdown.Both);	// Close the connection (in both directions)
            }
            catch
            {   // Generic approach to dealing with situations such as socket not connected
            }
        }

        void CloseConnectionAndSocket()
        {
            if(null != m_Timer_Application)
            {
                m_Timer_Application.Enabled = false;
            }
            CloseConnection();
            try
            {
                m_ConnectSock.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if (sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }
    }
}
