﻿/*	File			MediaShare_Peer.cs
	Purpose			Peer-peer demonstration
	Author			Richard John Anthony
	Date			February 2015
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace MediaSharePeer
{
    public partial class MediaShare_Peer : Form
    {
        public const Int16 PEER_DISCOVERY_PORT = 8004;	// Decimal (Receive broadcast self-advertisement messages from other peers)
        public const Int16 PEER_APPLICATION_PORT = 8005;// Decimal (Receive point-to-point UDP messages from other peers)

        public const Int32 MAX_KNOWN_PEERS = 10;        // Demonstration, therefore a fixed array of ten peers is adequate
        public const Int32 MAX_SONG_TITLE_LEN = 40;
        public const Int32 MAX_ARTIST_NAME_LEN = 30;
        public const Int32 MAX_DATA_ROWS_LOCAL = 4;
        public const Int32 MAX_DATA_ROWS = MAX_DATA_ROWS_LOCAL * MAX_KNOWN_PEERS;

        public enum MessageType {PEER_SELF_ADVERTISEMENT, REQUEST_MEDIA_LIST, SEND_MEDIA_RECORD};

        // Message structure (this will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Application_Message_PDU
        {
		    public MessageType eMessageType;    // Offset 0
	        public Int32 iSender_Random_ID;     // Offset 4
	        public bool bDataValid;             // Offset 8
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_SONG_TITLE_LEN + 1))]
	        public byte[] cSongTitle;	        // Offset 9 - 49 inclusive
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_ARTIST_NAME_LEN + 1))]
	        public byte[] cArtistName;          // Offset 50 - 80 inclusive
            // There is a 3-byte pad in the message as actually transmitted by the C++ version to truncate the message on a 4-byte boundary
        };

        public const Int16 Application_Message_Size = 84;   // For serialisation / deserialisation (for compatibility with C++ version)

        // Message structure (this will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct PeerSelfAdvertisement_Message_PDU
        {
	        public MessageType eMessageType;    // Offset 0
		    public Int32 iSender_Random_ID;     // Offset 4
			
		    // PeerAddressDetails_SockAddr;
            public Int16 iAddressFamily;        // Offset 8
            public Int16 iPort;                 // Offset 10-11
            public byte iIP_Address_Byte_1;	    // Offset 12
            public byte iIP_Address_Byte_2;     // Offset 13
            public byte iIP_Address_Byte_3;     // Offset 14
            public byte iIP_Address_Byte_4;	    // Offset 15
            // The C++ version includes a 16-byte Sockaddr_IN structure to hold the Address and port value (see serialisation / deserialisation code)
        };

        public const Int16 PeerSelfAdvertisement_Message_PDU_Size = 24;   // For serialisation / deserialisation (for compatibility with C++ version)

        public struct Known_Peer_details 
        {
	        public Int32 iPeer_ID;
            public IPEndPoint Peer_IPEndPoint;
            public bool bInUse;
	        public bool bMediaDataReceived;
        };

        // Data Row and Table structures (hard coded data serves as mock of real database containing media files (songs))
        public struct DataRow_LOCAL_struct 
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_SONG_TITLE_LEN + 1))]
	        public byte[] cSongTitle;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_ARTIST_NAME_LEN + 1))]
	        public byte[] cArtistName;
        };

        public struct DataRow_struct
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_SONG_TITLE_LEN + 1))]
            public byte[] cSongTitle;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_ARTIST_NAME_LEN + 1))]
            public byte[] cArtistName;
            public Int32 iProviderPeerID;
        };

	    public DataRow_LOCAL_struct[] m_LocalDataRowArray;
	    public DataRow_struct[] m_DataRowArray; // Array of data row structures
	    public Known_Peer_details[] m_Known_Peers_List;

	    Int32 m_iPeer_Random_ID;
	    Int32 m_iPeer_MediaConfiguration;

        Socket m_ReceivePeerDiscovery_Sock;
	    Socket m_ReceiveApplicationMessages_Sock;
	    Socket m_SelfAdvertisementBroadcast_Sock;
	    Socket m_SendApplicationMessages_Sock;	

 	    IPEndPoint m_Local_IPEndPoint_ForReceivePeerDiscovery;
 	    IPEndPoint m_Local_IPEndPoint_ForReceiveApplicationMessages;
 	    IPEndPoint m_IPEndPoint_ForSelfAdvertisement_BROADCAST;

        System.Windows.Forms.Timer m_Timer_Application;
        System.Windows.Forms.Timer m_Timer_Send_ServerAdvertisement;

        public MediaShare_Peer()
        {
            InitializeComponent();
 
            Random rnd = new Random();
 	        m_iPeer_Random_ID = rnd.Next(1,10000);
            m_iPeer_MediaConfiguration = rnd.Next(0, 10/*exclusive upper value*/); // Randomly select a media configuration betwen 0 and 9

	        Initialise_KnownPeers_Array();
	        InitialiseDataTable_LOCAL();
	        InitialiseDataTable_FULL();
            DisplayDataTable();

            GetLocalHostAddressDetails();
            InitialiseSocket_ForReceivePeerDiscoveryBroadcasts();
	        InitialiseSocket_ForReceiveApplicationMessages();
	        InitialiseSocket_ForSendSelfAdvertisement_BROADCAST();
            InitialiseSocket_SendApplicationMessages();

	        PeerID_textBox.Text = m_iPeer_Random_ID.ToString();
            PeerConfiguration_textBox.Text = m_iPeer_MediaConfiguration.ToString();
            EnablePeer();
        }

        void GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {   // Create the local IPEndPoint using the first IPv4 address and the pre-defined port
                    m_Local_IPEndPoint_ForReceivePeerDiscovery = new IPEndPoint(ipHostAddress[iAddrIndex], PEER_DISCOVERY_PORT);
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    string sStr = string.Format("Local Address (recv peer discovery messages): {0}.{1}.{2}.{3}  Port:{4}", 
                        bAddressBytes[0].ToString(), bAddressBytes[1].ToString(), bAddressBytes[2].ToString(),
                        bAddressBytes[3].ToString(), PEER_DISCOVERY_PORT);
                    WriteStatusLine(sStr);
 
                    m_Local_IPEndPoint_ForReceiveApplicationMessages = new IPEndPoint(ipHostAddress[iAddrIndex], PEER_APPLICATION_PORT);
                    bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    sStr = string.Format("Local Address (recv application messages): {0}.{1}.{2}.{3}  Port:{4}", 
                        bAddressBytes[0].ToString(), bAddressBytes[1].ToString(), bAddressBytes[2].ToString(),
                        bAddressBytes[3].ToString(), PEER_APPLICATION_PORT);
                    WriteStatusLine(sStr);
                    return;
                }
            }
            MessageBox.Show("Failed to get local address!", "MediaShare_Peer");
            CloseSocketAndExit();
        }
        
        void InitialiseAddressStructAndPort_For_BROADCAST_ServerAdvertisement()
        {	// Initialise the IP address
            m_IPEndPoint_ForSelfAdvertisement_BROADCAST = new IPEndPoint(IPAddress.Broadcast, PEER_DISCOVERY_PORT);
        }

        void Initialise_KnownPeers_Array()
        {
            m_Known_Peers_List = new Known_Peer_details[MAX_KNOWN_PEERS];

            for (int iIndex = 0; iIndex < MAX_KNOWN_PEERS; iIndex++)
	        {
		        m_Known_Peers_List[iIndex].bInUse = false;		
		        m_Known_Peers_List[iIndex].bMediaDataReceived = false;
	        }
        }

        Int32 GetIndexOf_Available_KnownPeersList_Entry()
        { 
	        int iNextAvailableEntryIndex = -1;	// Signals that the array is full
	        for(int iIndex = 0; iIndex < MAX_KNOWN_PEERS; iIndex++)
	        {
		        if(false == m_Known_Peers_List[iIndex].bInUse)
		        {
			        iNextAvailableEntryIndex = iIndex;
			        break;
		        }
	        }
	        return iNextAvailableEntryIndex;
        }

        Int32 GetIndexOf_KnownPeer_InPeersList(int iPeerID)
        {	
	        for(int iIndex = 0; iIndex < MAX_KNOWN_PEERS; iIndex++)
	        {
		        if(true == m_Known_Peers_List[iIndex].bInUse)
		        {
			        if(iPeerID == m_Known_Peers_List[iIndex].iPeer_ID)
			        {  // Matched by Peer ID
				        return iIndex;
			        }
		        }
	        }
	        return -1;
        }

        bool CheckIfPeerExistsIn_KnownPeersList(int iPeerID, IPEndPoint Peer_IPEndPoint)
        { 
	        for(int iIndex = 0; iIndex < MAX_KNOWN_PEERS; iIndex++)
	        {
		        if(true == m_Known_Peers_List[iIndex].bInUse)
		        {
			        if(iPeerID == m_Known_Peers_List[iIndex].iPeer_ID &&
                        m_Known_Peers_List[iIndex].Peer_IPEndPoint.ToString().Equals(Peer_IPEndPoint.ToString()))
			        {  // Matched by combination of ID and IP address
				        return true;
			        }
		        }
	        }
	        return false;
        }

        bool Add_NewlyDiscoveredPeerTo_KnownPeersList(int iPeer_ID, IPEndPoint Peer_IPEndPoint)
        {
	        int iIndex = GetIndexOf_Available_KnownPeersList_Entry();
	        if(-1 == iIndex)
	        {
		        return false;	// No space in array
	        }		
	        m_Known_Peers_List[iIndex].bInUse = true;
	        m_Known_Peers_List[iIndex].iPeer_ID = iPeer_ID;
            m_Known_Peers_List[iIndex].Peer_IPEndPoint = Peer_IPEndPoint;
	        return true;
        }

        void InitialiseDataTable_LOCAL()	// Populate  m_DataTable_LOCAL (hardcoded data)
        {
            m_LocalDataRowArray = new DataRow_LOCAL_struct[MAX_DATA_ROWS_LOCAL];

            for (int iIndex = 0; iIndex < MAX_DATA_ROWS_LOCAL; iIndex++)
	        {   // Initialise m_DataTable_LOCAL (all strings blank)
                m_LocalDataRowArray[iIndex].cSongTitle = new byte[MAX_SONG_TITLE_LEN + 1];
                m_LocalDataRowArray[iIndex].cArtistName = new byte[MAX_ARTIST_NAME_LEN + 1];
	        }
	
	        switch(m_iPeer_MediaConfiguration) // Values 0 - 9 chosen randomly to simulate different peer instances
	        {
		        case 0:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("Bicycle Race");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("Queen");
			        m_LocalDataRowArray[1].cSongTitle = StringTo_ASCII_ByteArray("Sir Duke");
			        m_LocalDataRowArray[1].cArtistName = StringTo_ASCII_ByteArray("Stevie Wonder");
			        break;
		        case 1:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("Undercover Of The Night");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("The Rolling Stones");
			        m_LocalDataRowArray[1].cSongTitle = StringTo_ASCII_ByteArray("Dancing In The Moonlight");
			        m_LocalDataRowArray[1].cArtistName = StringTo_ASCII_ByteArray("Thin Lizzy");
			        m_LocalDataRowArray[2].cSongTitle = StringTo_ASCII_ByteArray("Get Back");
			        m_LocalDataRowArray[2].cArtistName = StringTo_ASCII_ByteArray("The Beatles");
			        break;
		        case 2:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("Living For The City");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("Stevie Wonder");
			        break;
		        case 3:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("Dont Stop Me Now");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("Queen");
			        break;
		        case 4:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("Start Me UP");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("The Rolling Stones");
			        m_LocalDataRowArray[1].cSongTitle = StringTo_ASCII_ByteArray("My Generation");
			        m_LocalDataRowArray[1].cArtistName = StringTo_ASCII_ByteArray("The Who");
			        break;
		        case 5:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("Let It Be");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("The Beatles");
			        m_LocalDataRowArray[1].cSongTitle = StringTo_ASCII_ByteArray("Rock And Roll");
			        m_LocalDataRowArray[1].cArtistName = StringTo_ASCII_ByteArray("Led Zeppelin");
			        m_LocalDataRowArray[2].cSongTitle = StringTo_ASCII_ByteArray("Master Blaster");
			        m_LocalDataRowArray[2].cArtistName = StringTo_ASCII_ByteArray("Stevie Wonder");
			        m_LocalDataRowArray[3].cSongTitle = StringTo_ASCII_ByteArray("Leaving Here");
			        m_LocalDataRowArray[3].cArtistName = StringTo_ASCII_ByteArray("The Who");
			        break;
		        case 6:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("House Of The Rising Sun");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("The Animals");
			        m_LocalDataRowArray[1].cSongTitle = StringTo_ASCII_ByteArray("The Battle Of Evermore");
			        m_LocalDataRowArray[1].cArtistName = StringTo_ASCII_ByteArray("Led Zeppelin");
			        m_LocalDataRowArray[2].cSongTitle = StringTo_ASCII_ByteArray("Sarah");
			        m_LocalDataRowArray[2].cArtistName = StringTo_ASCII_ByteArray("Thin Lizzy");
			        break;
		        case 7:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("Whole Lotta Love");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("Led Zeppelin");
			        m_LocalDataRowArray[1].cSongTitle = StringTo_ASCII_ByteArray("We've Gotta Get Out Of This Place");
			        m_LocalDataRowArray[1].cArtistName = StringTo_ASCII_ByteArray("The Animals");
			        m_LocalDataRowArray[2].cSongTitle = StringTo_ASCII_ByteArray("Seven Seas Of Rhye");
			        m_LocalDataRowArray[2].cArtistName = StringTo_ASCII_ByteArray("Queen");
			        m_LocalDataRowArray[3].cSongTitle = StringTo_ASCII_ByteArray("Ticket To Ride");
			        m_LocalDataRowArray[3].cArtistName = StringTo_ASCII_ByteArray("The Beatles");
			        break;
		        case 8:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("Hey Jude");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("The Beatles");
			        break;
		        case 9:
			        m_LocalDataRowArray[0].cSongTitle = StringTo_ASCII_ByteArray("Waiting For An Alibi");
			        m_LocalDataRowArray[0].cArtistName = StringTo_ASCII_ByteArray("Thin Lizzy");
			        m_LocalDataRowArray[1].cSongTitle = StringTo_ASCII_ByteArray("Harlem Shuffle");
			        m_LocalDataRowArray[1].cArtistName = StringTo_ASCII_ByteArray("The Rolling Stones");
			        break;
	        }
        }

        void InitialiseDataTable_FULL()	// Upate DataTable
        {
            m_DataRowArray = new DataRow_struct[MAX_DATA_ROWS];
            
            int iIndex;
	        for(iIndex = 0; iIndex < MAX_DATA_ROWS; iIndex++)
	        {   // Initialise DataTable (all strings blank)
                m_DataRowArray[iIndex].cSongTitle = new byte[MAX_SONG_TITLE_LEN + 1];
                m_DataRowArray[iIndex].cArtistName = new byte[MAX_ARTIST_NAME_LEN + 1];
                m_DataRowArray[iIndex].iProviderPeerID = -1;
	        }

            // Copy local data rows into full data table, until there are no more valid entries, or the maximum number of local entries is reached
            for(iIndex = 0; iIndex < MAX_DATA_ROWS_LOCAL; iIndex++)
	        {
                if (0 != m_LocalDataRowArray[iIndex].cSongTitle[0])
                {
                    Buffer.BlockCopy(m_LocalDataRowArray[iIndex].cSongTitle/*src*/, 0, m_DataRowArray[iIndex].cSongTitle/*dest*/, 0, m_LocalDataRowArray[iIndex].cSongTitle.Length);
                    Buffer.BlockCopy(m_LocalDataRowArray[iIndex].cArtistName/*src*/, 0, m_DataRowArray[iIndex].cArtistName/*dest*/, 0, m_LocalDataRowArray[iIndex].cArtistName.Length);
                    m_DataRowArray[iIndex].iProviderPeerID = m_iPeer_Random_ID;
                }
	        }
        }

        void Add_Row_To_DataTable_FULL(byte[] cSongTitle, byte[] cArtistName, Int32 iProviderPeerID)	// Upate DataTable
        {
	        int iIndex;
	        iIndex = 0;
            while (0 != m_DataRowArray[iIndex].cSongTitle[0])
	        {   // Skip populated entries
		        iIndex++;
	        }
	        // iIndex now points to next available entry
	        if(iIndex < MAX_DATA_ROWS)
	        {
                Buffer.BlockCopy(cSongTitle/*src*/, 0, m_DataRowArray[iIndex].cSongTitle/*dest*/, 0, MAX_SONG_TITLE_LEN);
                Buffer.BlockCopy(cArtistName/*src*/, 0, m_DataRowArray[iIndex].cArtistName/*dest*/, 0, MAX_ARTIST_NAME_LEN);
                m_DataRowArray[iIndex].iProviderPeerID = iProviderPeerID;
	        }
	        DisplayDataTable();
        }

        void DisplayDataTable()
        {	
            Song_listBox.Items.Clear();
            Artist_listBox.Items.Clear();
            ProviderPeer_listBox.Items.Clear();
            string sStr = "";
	        for(int iIndex = 0; iIndex < MAX_DATA_ROWS; iIndex++)
	        {   // Display all valid rows (determined by the song title field not being the empty string "")
                if (0 != m_DataRowArray[iIndex].cSongTitle[0])
		        {
                    ByteArrayTo_ASCII_String(m_DataRowArray[iIndex].cSongTitle, ref sStr);
                    Song_listBox.Items.Add(sStr);
                    ByteArrayTo_ASCII_String(m_DataRowArray[iIndex].cArtistName, ref sStr);
                    Artist_listBox.Items.Add(sStr);
                    ProviderPeer_listBox.Items.Add(m_DataRowArray[iIndex].iProviderPeerID.ToString());
		        }
	        }
        }

        void InitialiseSocket_ForReceivePeerDiscoveryBroadcasts()
        {	
	        // Create Socket and set it NonBlocking
	        try
            {
                m_ReceivePeerDiscovery_Sock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
	            m_ReceivePeerDiscovery_Sock.Blocking = false;
            }
            catch
            {
		        MessageBox.Show("Socket create failed","MediaShare Peer");
		        CloseSocketAndExit();	
	        }
	
	        try
            {
                m_ReceivePeerDiscovery_Sock.Bind(m_Local_IPEndPoint_ForReceivePeerDiscovery);
            }
            catch
            {
		        MessageBox.Show("Bind failed !\nPort is probably in use","MediaShare Peer");
                CloseSocketAndExit();
            }
        }

        void InitialiseSocket_ForReceiveApplicationMessages()
        {		
	        // Create Socket and set it NonBlocking
	        try
            {
                m_ReceiveApplicationMessages_Sock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
	            m_ReceiveApplicationMessages_Sock.Blocking = false;
            }
            catch
            {
		        MessageBox.Show("Socket create failed","MediaShare Peer");
		        CloseSocketAndExit();	
	        }
	
	        try
            {
                m_ReceiveApplicationMessages_Sock.Bind(m_Local_IPEndPoint_ForReceiveApplicationMessages);
            }
            catch
            {
		        MessageBox.Show("Bind failed !\nPort is probably in use","MediaShare Peer");
                CloseSocketAndExit();
            }
        }

        void InitialiseSocket_ForSendSelfAdvertisement_BROADCAST()
        {	
	        // Create Socket and enable broadcast
	        try
            {
                m_SelfAdvertisementBroadcast_Sock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
	            m_SelfAdvertisementBroadcast_Sock.EnableBroadcast = true;
            }
            catch
            {
		        MessageBox.Show("Socket create failed","MediaShare Peer");
		        CloseSocketAndExit();	
	        }

                m_IPEndPoint_ForSelfAdvertisement_BROADCAST = new IPEndPoint(IPAddress.Broadcast, PEER_DISCOVERY_PORT);
        }

        void InitialiseSocket_SendApplicationMessages()
        {	// Create Socket (sends to specific addressed peers, does not broadcast)
            try
            {
                m_SendApplicationMessages_Sock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch
            {
                MessageBox.Show("Socket create failed", "MediaShare Peer");
                CloseSocketAndExit();
            }
        }

        void EnablePeer()  // Invoked on startup
        {
	        Start_Timer_Receive();   // As all activity is non-blocking
            Start_Timer_Send_ServerAdvertisement();
        }

        void Start_Timer_Receive()
        {
            m_Timer_Application = new System.Windows.Forms.Timer();
            m_Timer_Application.Interval = 500; // milliseconds
            m_Timer_Application.Tick += OnTimedEvent_Timer_Receive;
            m_Timer_Application.Enabled = true;
        }

        void OnTimedEvent_Timer_Receive(Object myObject, EventArgs myEventArgs)
        {
			DoReceive_ApplicationMessages();	// Check for application messages from known peers
			DoReceive_PeerAdvertisements();		// Check for self advertisement peers and discover previously unknown peers
        }

        void Start_Timer_Send_ServerAdvertisement()
        {
            m_Timer_Send_ServerAdvertisement = new System.Windows.Forms.Timer();
            m_Timer_Send_ServerAdvertisement.Interval = 5000;	// 5 second interval between sending ServerAdvertisement messages
            m_Timer_Send_ServerAdvertisement.Tick += OnTimedEvent_Timer_Send_ServerAdvertisement;
            m_Timer_Send_ServerAdvertisement.Enabled = true;
        }

        void OnTimedEvent_Timer_Send_ServerAdvertisement(Object myObject, EventArgs myEventArgs)
        {
			Send_SelfAdvertisement();  // Broadcast Self Advertisement message to other Peers
			Send_REQUEST_MEDIA_LIST_To_NewlyDiscoveredPeers(); 
        }

        void Initialise_Send_ServerAdvertisement_Socket()
        {
            try
            {
                m_SelfAdvertisementBroadcast_Sock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_SelfAdvertisementBroadcast_Sock.EnableBroadcast = true;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "MediaShare_Peer");
                CloseSocketAndExit();
            }
	        InitialiseAddressStructAndPort_For_BROADCAST_ServerAdvertisement();
        }

        void Send_SelfAdvertisement()
        {
            PeerSelfAdvertisement_Message_PDU SelfAdvertisement_Message = new PeerSelfAdvertisement_Message_PDU();
            SelfAdvertisement_Message.eMessageType = MessageType.PEER_SELF_ADVERTISEMENT;
            SelfAdvertisement_Message.iSender_Random_ID = m_iPeer_Random_ID;

            byte[] bAddress = m_Local_IPEndPoint_ForReceiveApplicationMessages.Address.GetAddressBytes();
            SelfAdvertisement_Message.iIP_Address_Byte_1 = bAddress[0];
            SelfAdvertisement_Message.iIP_Address_Byte_2 = bAddress[1];
            SelfAdvertisement_Message.iIP_Address_Byte_3 = bAddress[2];
            SelfAdvertisement_Message.iIP_Address_Byte_4 = bAddress[3];
            SelfAdvertisement_Message.iPort = PEER_APPLICATION_PORT;
            SelfAdvertisement_Message.iAddressFamily = 2;   // For C++ compatibility, value 2 = AF_INET

            byte[] bBuffer = Serialize_PeerSelfAdvertisement_Message_PDU(SelfAdvertisement_Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_SelfAdvertisementBroadcast_Sock.SendTo(bBuffer, m_IPEndPoint_ForSelfAdvertisement_BROADCAST);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Place diagnostics here if necessay
                }
            }
        }

        void SendApplicationMessage(Application_Message_PDU Message, IPEndPoint To_IPEndPoint)
        {
            // The sendto addresss is set dynamically (depending on which known Peer the application message is to be sent to)
            byte[] bBuffer = Serialize_Application_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_SendApplicationMessages_Sock.SendTo(bBuffer, To_IPEndPoint);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                }
            }
        }

        void Send_REQUEST_MEDIA_LIST_To_NewlyDiscoveredPeers()
        {   // A race can occur if an application message (especially the first one sent: REQUEST_MEDIA_LIST) is sent
	        // immediately after a remote peer is discovered. This is because discovery is two-sided and the fact that
	        // the local peer has discovered the remote peer does not guarantee that the remote peer has discovered the 
	        // local peer. Received application messages that do not originate from known peers (i.e. in local m_Known_Peers_List)
	        // are ignored, hence the need for a special means of issueing REQUEST_MEDIA_LIST messages (which has the effect of 
	        // decoupling peer discovery and requesting Media data):
	        // - For each discovered peer an entry is made in the m_Known_Peers_List array. 
	        // - For each entry the bMediaDataReceived flag is initially false, only set true once the Media data has been received.
	        // - Each time this method is invoked (via the timer) the m_Known_Peers_List is scanned to see if there are any new
	        // peers for which a REQUEST_MEDIA_LIST must be issued.
	        for(int iIndex = 0; iIndex < MAX_KNOWN_PEERS; iIndex++)
	        {
		        if(true == m_Known_Peers_List[iIndex].bInUse)
		        {
			        if(false == m_Known_Peers_List[iIndex].bMediaDataReceived)
			        {  // Do not yet have Media data; request it now
				        DoSend_REQUEST_MEDIA_LIST_To_NewlyDiscoveredPeer(m_Known_Peers_List[iIndex].iPeer_ID);
			        }
		        }
	        }
        }

        void DoSend_REQUEST_MEDIA_LIST_To_NewlyDiscoveredPeer(int iNewlyDiscoveredPeerID)
        {
	        int iIndex = GetIndexOf_KnownPeer_InPeersList(iNewlyDiscoveredPeerID);
	        if(-1 == iIndex)
	        {
		        WriteStatusLine("Target peer ID not found in Known Peers array"); // Should not be able to happen 
		        return;
	        }

            Application_Message_PDU Message = new Application_Message_PDU();
	        Message.eMessageType = MessageType.REQUEST_MEDIA_LIST;
	        Message.bDataValid = false;
	        Message.iSender_Random_ID = m_iPeer_Random_ID;
            SendApplicationMessage(Message, m_Known_Peers_List[iIndex].Peer_IPEndPoint);

            byte[] bAddress = m_Known_Peers_List[iIndex].Peer_IPEndPoint.Address.GetAddressBytes();
            Int16 iPort = (Int16)m_Known_Peers_List[iIndex].Peer_IPEndPoint.Port;
	        string sStr = string.Format("REQUEST_MEDIA_LIST request sent to IP Address:{0}.{1}.{2}.{3}  Port:{4}",
                                                      bAddress[0], bAddress[1], bAddress[2], bAddress[3], iPort);
	        WriteStatusLine(sStr);
        }

        void Send_MediaFilesDetails_To_requestor(IPEndPoint To_IPEndPoint)
        {
	        Application_Message_PDU Message = new Application_Message_PDU();
	        Message.eMessageType = MessageType.SEND_MEDIA_RECORD;
	        Message.iSender_Random_ID = m_iPeer_Random_ID;
	        Message.bDataValid = true;
            Message.cSongTitle = new byte[MAX_SONG_TITLE_LEN + 1];
            Message.cArtistName = new byte[MAX_ARTIST_NAME_LEN + 1];
	
	        for(int iIndex = 0; iIndex < MAX_DATA_ROWS_LOCAL; iIndex++)
	        {   // Send all valid rows, one by one, (determined by the song title field not being the empty string "")
		        // Send only the data concerning media files held locally
		        if(0 != m_LocalDataRowArray[iIndex].cSongTitle[0])
		        {
                    Buffer.BlockCopy(m_LocalDataRowArray[iIndex].cSongTitle/*src*/, 0, Message.cSongTitle/*dest*/, 0, m_LocalDataRowArray[iIndex].cSongTitle.Length);
                    Buffer.BlockCopy(m_LocalDataRowArray[iIndex].cArtistName/*src*/, 0, Message.cArtistName/*dest*/, 0, m_LocalDataRowArray[iIndex].cArtistName.Length);
                    SendApplicationMessage(Message, To_IPEndPoint);
                    byte[] bAddress = To_IPEndPoint.Address.GetAddressBytes();
                    Int16 iPort = (Int16)To_IPEndPoint.Port;
                    string sStr = string.Format("SEND_MEDIA_RECORD sent to IP Address:{0}.{1}.{2}.{3}  Port:{4}",
                                                    bAddress[0], bAddress[1], bAddress[2], bAddress[3], iPort);
                    WriteStatusLine(sStr);
		        }
	        }			
	        Message.cSongTitle = StringTo_ASCII_ByteArray("--END--"); // Signals end of data
            SendApplicationMessage(Message, To_IPEndPoint);
        }

        void DoReceive_ApplicationMessages()
        {
            // Check for Application messages (from ANY other peers)
            byte[] bArray = new byte[Application_Message_Size];
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;
            iReceiveByteCount = 0;
            try
            {
                iReceiveByteCount = m_ReceiveApplicationMessages_Sock.ReceiveFrom(bArray, ref From_EndPoint);
                From_IPEndPoint = ((IPEndPoint)From_EndPoint); // Store sender's IPEndPoint
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Additional diagnostics can be placed here if necessary
                }
                return;
            }
            if (0 < iReceiveByteCount)
            {
                Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
 
  		        Int32 iIndex = GetIndexOf_KnownPeer_InPeersList(Message.iSender_Random_ID);
		        if(-1 == iIndex)
		        {
			        WriteStatusLine("Sender ID not found in Known Peers array");
			        return;
		        }
          
 		        // If here; found address of sender to display, and for replying to, in Known Peers array
                byte[] bAddress = From_IPEndPoint.Address.GetAddressBytes();
                Int16 iPort = (Int16)From_IPEndPoint.Port;

                string sStr = "";
                switch(Message.eMessageType)
		        {
			        case MessageType.REQUEST_MEDIA_LIST:
                        sStr = string.Format("REQUEST_MEDIA_LIST received, ID:{0}  IP Address:{1}.{2}.{3}.{4}  Port:{5}",
					                    Message.iSender_Random_ID, bAddress[0], bAddress[1], bAddress[2], bAddress[3], iPort);
				        WriteStatusLine(sStr);
				        Send_MediaFilesDetails_To_requestor(m_Known_Peers_List[iIndex].Peer_IPEndPoint);
				        break;
	
			        case MessageType.SEND_MEDIA_RECORD:
                        ByteArrayTo_ASCII_String(Message.cSongTitle, ref sStr);
				        if(true == sStr.Contains("--END--"))
				        {  // Signals that all media data has been sent by the remote peer
					        m_Known_Peers_List[iIndex].bMediaDataReceived = true; 
					        // Prevents further REQUEST_MEDIA_LIST messages being sent to the particular remote peer
				        }
				        else
				        {	// Update the data table and display
                            if (false == m_Known_Peers_List[iIndex].bMediaDataReceived)
                            {
                                sStr = string.Format("SEND_MEDIA_RECORD received, ID:{0}  IP Address:{1}.{2}.{3}.{4}  Port:{5}",
                                        Message.iSender_Random_ID, bAddress[0], bAddress[1], bAddress[2], bAddress[3], iPort);
                                WriteStatusLine(sStr);
                                sStr = string.Format("MEDIA received, Song:{0}  Artist:{1}",
                                    ByteArrayTo_NullTerminated_ASCII_String(Message.cSongTitle), 
                                    ByteArrayTo_NullTerminated_ASCII_String(Message.cArtistName));
                                WriteStatusLine(sStr);
                                Add_Row_To_DataTable_FULL(Message.cSongTitle, Message.cArtistName, Message.iSender_Random_ID);
                            }
				        }
				        break;
                }
            }
        }

        void DoReceive_PeerAdvertisements()
        {		
	        // Check for self-advertisement messages from other peers (UDP broadcast from any peer instance, need to ignore messages from self)
            byte[] bArray = new byte[PeerSelfAdvertisement_Message_PDU_Size];
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;
            iReceiveByteCount = 0;
            try
            {
                iReceiveByteCount = m_ReceivePeerDiscovery_Sock.ReceiveFrom(bArray, ref From_EndPoint);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Additional diagnostics can be placed here if necessary
                }
                return;
            }
            if (0 < iReceiveByteCount)
            {
                PeerSelfAdvertisement_Message_PDU PeerSelfAdvertisement_Message = DeSerialize_PeerSelfAdvertisementMessage(bArray);
                byte[] bAddress = new byte[4];
                bAddress[0] = PeerSelfAdvertisement_Message.iIP_Address_Byte_1;
                bAddress[1] = PeerSelfAdvertisement_Message.iIP_Address_Byte_2;
                bAddress[2] = PeerSelfAdvertisement_Message.iIP_Address_Byte_3;
                bAddress[3] = PeerSelfAdvertisement_Message.iIP_Address_Byte_4;
                IPAddress SenderIPAddress = new IPAddress(bAddress);
                IPEndPoint SenderIPEndPoint = new IPEndPoint(SenderIPAddress, PeerSelfAdvertisement_Message.iPort);

                switch (PeerSelfAdvertisement_Message.eMessageType)
                {
				    case MessageType.PEER_SELF_ADVERTISEMENT:
                        if (false == m_Local_IPEndPoint_ForReceiveApplicationMessages.ToString().Equals(SenderIPEndPoint.ToString()))
					    {   // Check if message is from self - if so ignore

						    if(false == CheckIfPeerExistsIn_KnownPeersList(PeerSelfAdvertisement_Message.iSender_Random_ID,
                                                                                                            SenderIPEndPoint))
						    {   // Check if peer is already known (need combination of ID and address), Here if not yet known
					
							    if(true == Add_NewlyDiscoveredPeerTo_KnownPeersList(PeerSelfAdvertisement_Message.iSender_Random_ID,
																		                                    SenderIPEndPoint))
							    {
                                    string sStr = string.Format("New peer discovered, ID:{0}  IP Address:{1}.{2}.{3}.{4}  Port:{5}",
                                        PeerSelfAdvertisement_Message.iSender_Random_ID, bAddress[0], bAddress[1], bAddress[2], 
                                                                                bAddress[3], PeerSelfAdvertisement_Message.iPort);
                                    WriteStatusLine(sStr);
							    }
							    else
							    {
								    string sStr = string.Format("New peer discovered ID:{0} !No space in KnownPeers array!",
									    PeerSelfAdvertisement_Message.iSender_Random_ID);
								    WriteStatusLine(sStr);							 
							    }
						    }
					    }
				    break;
			    }
		    }
	    }
 
        private void ClearLog_button_Click(object sender, EventArgs e)
        {
            EventLog_listBox.Items.Clear();
        }
        
        PeerSelfAdvertisement_Message_PDU DeSerialize_PeerSelfAdvertisementMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a ServerAdvertisement_Message_PDU structure
            PeerSelfAdvertisement_Message_PDU result_Message = new PeerSelfAdvertisement_Message_PDU();
            //int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0 /*FieldOffset*/);
            result_Message.iSender_Random_ID = ByteArrayToInt32(bRecdByteArray, 4 /*FieldOffset*/);
            // Deserialisation of port number
            // The message format and deserialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            result_Message.iAddressFamily = bRecdByteArray[8]; // Value 2 = AF_INET
            result_Message.iPort = (Int16)((256 * bRecdByteArray[10])/*1st of the 4 bytes*/ + bRecdByteArray[11]/*2nd of the 4 bytes*/);
            result_Message.iIP_Address_Byte_1 = bRecdByteArray[12];
            result_Message.iIP_Address_Byte_2 = bRecdByteArray[13];
            result_Message.iIP_Address_Byte_3 = bRecdByteArray[14];
            result_Message.iIP_Address_Byte_4 = bRecdByteArray[15];
            return result_Message;
        }

        Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
            Application_Message_PDU result_Message = new Application_Message_PDU();
            //int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0/*FieldOffset*/);
            result_Message.iSender_Random_ID = ByteArrayToInt32(bRecdByteArray, 4/*FieldOffset*/);
            result_Message.bDataValid = true;
            if (0 == bRecdByteArray[8])
            {
                result_Message.bDataValid = false;
            }
            result_Message.cSongTitle = new byte[MAX_SONG_TITLE_LEN + 1];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 9, result_Message.cSongTitle/*dest*/, 0, MAX_SONG_TITLE_LEN/*count*/);
            result_Message.cArtistName = new byte[MAX_ARTIST_NAME_LEN + 1];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 50, result_Message.cArtistName/*dest*/, 0, MAX_ARTIST_NAME_LEN/*count*/);
            return result_Message;
        }
     
        private byte[] Serialize_PeerSelfAdvertisement_Message_PDU(PeerSelfAdvertisement_Message_PDU Message)
        {   // Convert an ServerAdvertisement_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[PeerSelfAdvertisement_Message_PDU_Size];        // Create a byte array to hold the serialised form of the message PDU
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 0/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iSender_Random_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 4/*bByteArray offset*/, 4/*count*/);
            bByteArray[8] = (byte)Message.iAddressFamily;   // For C++ compatibility, value 2 = AF_INET
            // Serialisation of port number
            // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            bByteArray[10] = (byte)(Message.iPort / 256);
            bByteArray[11] = (byte)(Message.iPort % 256);
            bByteArray[12] = Message.iIP_Address_Byte_1;
            bByteArray[13] = Message.iIP_Address_Byte_2;
            bByteArray[14] = Message.iIP_Address_Byte_3;
            bByteArray[15] = Message.iIP_Address_Byte_4;
            return bByteArray;
        }

        private byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
        {   // Convert an Application_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[Application_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 0/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iSender_Random_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 4/*bByteArray offset*/, 4/*count*/);
            bByteArray[8] = 0; // false (bDataValid)
            if(true == Message.bDataValid)
            {
                bByteArray[8] = 1; // true (bDataValid)
            }
            if (null != Message.cSongTitle)
            {
                Buffer.BlockCopy(Message.cSongTitle/*src*/, 0, bByteArray/*dest*/, 9, Message.cSongTitle.Length/*count*/);
            }
            if (null != Message.cArtistName)
            {
                Buffer.BlockCopy(Message.cArtistName/*src*/, 0, bByteArray/*dest*/, 50, Message.cArtistName.Length/*count*/);
            }
            return bByteArray;
        }

        void WriteStatusLine(string sLine)
        {
            EventLog_listBox.Items.Add(sLine);
        }

        private void Done_Button_Click(object sender, EventArgs e)
        {
            CloseSocketAndExit();
        }

        void CloseSocketAndExit()
        {
            CloseSocket();
            Application.Exit();
        }
 
        void CloseSocket()
        {
            m_Timer_Application.Enabled = false;
            try
            {
                m_ReceivePeerDiscovery_Sock.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
            try
            {
                m_ReceiveApplicationMessages_Sock.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
            try
            {
                m_SelfAdvertisementBroadcast_Sock.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
            try
            {
                m_SendApplicationMessages_Sock.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }

        static string ByteArrayTo_NullTerminated_ASCII_String(byte[] bArray) // Returns string terminated at the null terminator
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            string sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            string sFinal = sStr.Substring(0,iASCIIz_length);
            return sFinal;
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if (sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }
    }
}
