﻿/*	File			Replication_Client.cs
	Purpose			Replication demonstration
	Author			Richard John Anthony
	Date			February 2015
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace ReplicationClient
{
    public partial class Replication_Client : Form
    {
        public const Int16 REPLICATION_EXTERNAL_PORT = 8006;  // Decimal (Client to Server)
        public const Int16 REPLICATION_UPDATE_PORT = 8007;	// Decimal (Server to Server - Update propagation)
        public const Int16 REPLICATION_HEARTBEAT_PORT = 8008; // Decimal (Server to Server - Master heartbeat broadcasts)
        public const Int16 REPLICATION_SERVER_ADVERTISEMENT_PORT = 8009; // Decimal (Server to client)

        public const Int16 MAX_CUST_NAME_LEN = 20;

        public enum ServerState {MASTER, BACKUP};
	        // Default state for server instances is BACKUP, in this state:
	        //                  - listen for Master heartbeats, and update propagations
	        //                  - do not accept client connections
	        //                  - elevate to MASTER state if three consecutive master heartbeats missed

	        // MASTER state, in this state:
	        //                  - send periodic Master heartbeats (broadcast)
	        //                  - listen for Master heartbeats
	        //                  - accept client connections
	        //                  - demote to BACKUP state if master heartbeat from another server instance is detected

        public enum MessageType {SERVER_ADVERTISEMENT, HEARTBEAT, REQUEST_RECORD, SEND_RECORD, 
            DATA_UPDATE_FROM_CLIENT, UPDATE_PROPAGATION_FROM_MASTER, CLIENT_DISCONNECT};

        // Message structures (these will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Application_Message_PDU 
        {
	        public MessageType eMessageType;
	        public int iSender_Random_ID;				// Needed so that UPDATE_PROPAGATION_FROM_MASTER messages from self can be detected and ignored
	        public int iCust_ID;
	        public bool bCustNameValid;                 // Single byte but will need to pad to 4-byte boundary (i.e. add 3 bytes padding)
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_CUST_NAME_LEN + 1))]
	        public byte[] cCustName;	
        };

        public const Int16 Application_Message_Size = 36;

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Heartbeat_Message_PDU
        {
	        public MessageType eMessageType;
	        public int iSender_Random_ID;				// Needed so that messages from self can be detected and ignored
	        public int iSender_Random_ID_Incremented;	// Used to differentiate genuine heartbeat messages from any other messages that just happen to arrive on this port
        };

        public const Int16 Heartbeat_Message_Size = 12;

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct ServerAdvertisement_Message_PDU
        {
	        public MessageType eMessageType;
	        public Int32 iSender_Random_ID;				
            public Int32 iAddressFamily;
            public Int32 iPort; // Port values are actually Int16, but encoded as Int32 in this message format for compatibility with the pre-existing C++ version
            public byte iIP_Address_Byte_1; 
            public byte iIP_Address_Byte_2; 
            public byte iIP_Address_Byte_3; 
            public byte iIP_Address_Byte_4; 
            public Int32 iPad; // Brings it up to 32 bytes
        };

        public const Int16 ServerAdvertisement_Message_Size = 24;

	    Socket m_LocalSocket_ForReceivingServerAdvertisements;
	    Socket m_ConnectSocket;
	    IPEndPoint m_IPEndPoint_ReceiveServerAdvertisements;
	    IPEndPoint m_IPEndPoint_ConnectToAppServer;

	    bool m_bConnected_AppServer;

        System.Windows.Forms.Timer m_Timer;
   
        public Replication_Client()
        {
            InitializeComponent();
	        InitialiseControls();
            GetLocalHostAddressDetails();
	        Enable_ReceiveServiceAdvertisementMessages();
        }

         void InitialiseControls()
        {
            Port_textBox.Enabled = false;
            ConnectAddress_1_TextBox.Enabled = false;
            ConnectAddress_2_TextBox.Enabled = false;
            ConnectAddress_3_TextBox.Enabled = false;
            ConnectAddress_4_TextBox.Enabled = false;
            Port_textBox.Text = "-";
            ConnectAddress_1_TextBox.Text = "-";
            ConnectAddress_2_TextBox.Text = "-";
            ConnectAddress_3_TextBox.Text = "-";
            ConnectAddress_4_TextBox.Text = "-";
             
            ConnectButton.Enabled = false;		// Do not enable the connect button until a service advertisement message has been received
	        m_bConnected_AppServer = false;	
            RequestData_button.Enabled = false;
            SendUpdatedData_button.Enabled = false;
 
            WriteStatusLine("Waiting for Server Advertisement message");
        }

        void GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            bool bIPv4_AddressFound = false;
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {   // Create the local IPEndPoint using the first IPv4 address and the pre-defined port
                    m_IPEndPoint_ReceiveServerAdvertisements = 
                        new IPEndPoint(ipHostAddress[iAddrIndex], REPLICATION_SERVER_ADVERTISEMENT_PORT);

                    // Initially display the local address (for diagnostic purposes). This is subsequently
                    // overwritten with the server address when it is received via a Server Advertisement message
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    ConnectAddress_1_TextBox.Text = bAddressBytes[0].ToString();
                    ConnectAddress_2_TextBox.Text = bAddressBytes[1].ToString();
                    ConnectAddress_3_TextBox.Text = bAddressBytes[2].ToString();
                    ConnectAddress_4_TextBox.Text = bAddressBytes[3].ToString();
                    Port_textBox.Text = REPLICATION_SERVER_ADVERTISEMENT_PORT.ToString();
                    bIPv4_AddressFound = true;
 
                    string sStr = string.Format("Client Address (for recv service advertisement): {0}.{1}.{2}.{3}  Port:{4}", 
                        bAddressBytes[0].ToString(), bAddressBytes[1].ToString(), bAddressBytes[2].ToString(),
                        bAddressBytes[3].ToString(), REPLICATION_SERVER_ADVERTISEMENT_PORT);
                    WriteStatusLine(sStr);
                    break;
                }
            }
            if (false == bIPv4_AddressFound)
            {   // No IPv4 address found - local host will not be able to connect to server using IPv4
                ConnectAddress_1_TextBox.Text = "No";
                ConnectAddress_2_TextBox.Text = "IP";
                ConnectAddress_3_TextBox.Text = "v4";
                ConnectAddress_4_TextBox.Text = "Addr";
                ConnectButton.Enabled = false;
            }
        }
        
        void GetServerAddress_FromServerAdvertisementMessage(ServerAdvertisement_Message_PDU ReceivedMessage)
        {	
	        // Extract the IP address and port number from the received Server Advertisement Message
            byte[] bAddress = new byte[4];
            bAddress[0] = ReceivedMessage.iIP_Address_Byte_1;
            bAddress[1] = ReceivedMessage.iIP_Address_Byte_2;
            bAddress[2] = ReceivedMessage.iIP_Address_Byte_3;
            bAddress[3] = ReceivedMessage.iIP_Address_Byte_4;
            IPAddress Address = new IPAddress(bAddress); // = 
            Int16 iPort = (Int16) ReceivedMessage.iPort;
            m_IPEndPoint_ConnectToAppServer = new IPEndPoint(Address, iPort);

            ConnectAddress_1_TextBox.Text = bAddress[0].ToString();
            ConnectAddress_2_TextBox.Text = bAddress[1].ToString();
            ConnectAddress_3_TextBox.Text = bAddress[2].ToString();
            ConnectAddress_4_TextBox.Text = bAddress[3].ToString();
            Port_textBox.Text = iPort.ToString();
	
            string sStr = string.Format("Received Server Advertisement IP: {0}.{1}.{2}.{3}  Port:{4}", 
                                bAddress[0].ToString(), bAddress[1].ToString(), bAddress[2].ToString(),
                                bAddress[3].ToString(), iPort);
            WriteStatusLine(sStr);
        }

	    void Create_ReceiveServerAdvertisementsSocket_AndSetItNonBlocking()
        {
            try
            {
                m_LocalSocket_ForReceivingServerAdvertisements = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_LocalSocket_ForReceivingServerAdvertisements.Blocking = false;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Replication Client");
                CloseSocketAndExit();
            }
        }

        void Enable_ReceiveServiceAdvertisementMessages()  // Invoked on startup
        {
	        Create_ReceiveServerAdvertisementsSocket_AndSetItNonBlocking();
            // Local address has already been retrived in GetLocalHostAddressDetails()
            try
            {
                m_LocalSocket_ForReceivingServerAdvertisements.Bind(m_IPEndPoint_ReceiveServerAdvertisements);
		        WriteStatusLine("Bind succeeded (Server Advertisement port)");
		        StartTimer();   // As all activity is non-blocking
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Bind failed (Server Advertisement port). Is Port in use by another instance?", "Replication Client");
                CloseSocketAndExit();
            }
        }

        void CreateConnectSocketAndSetBlocking()
        {
            try
            {
                m_ConnectSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ConnectSocket.Blocking = true; // Set blocking mode for socket during connection set-up phase
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Replication Client");
                CloseSocketAndExit();
            }
        }

        private void ConnectButton_Click(object sender, EventArgs e)
        {	// This button is not enabled until after a Service Advertisement mesage has been received
            // Therefore, by the time we get here the server's IP addrerss and port number are known to us
            // in the form of the IPEndPoint m_IPEndPoint_ConnectToAppServer
            CreateConnectSocketAndSetBlocking();
	
            try
            {
                m_ConnectSocket.Connect(m_IPEndPoint_ConnectToAppServer);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Connect failed !\n\nIs server address correct ?\nIs port correct ?\nIs server running ?", "Replication Client");
            }

	        m_bConnected_AppServer = true;
	        WriteStatusLine("Connected to Server");

            ConnectButton.Enabled= false;
            RequestData_button.Enabled = true;

	        // Set non-blocking (now that connection established) for receiving
            try
            {
                m_ConnectSocket.Blocking = false;
            }
            catch(SocketException)
		    {
                MessageBox.Show("Could not set non-Blocking behaviour for socket","Replication Client");
            }
	    }

        void StartTimer()
        {
            m_Timer = new System.Windows.Forms.Timer();
            m_Timer.Interval = 100; // milliseconds
            m_Timer.Tick += OnTimedEvent_Timer;
            m_Timer.Enabled = true;
        }

        void StopTimer()
        {
            if (null != m_Timer)
            {
                m_Timer.Enabled = false;
            }
        }

        void OnTimedEvent_Timer(Object myObject, EventArgs myEventArgs)
        {
            try
            {
                DoReceive();
            }
            catch // Silently catch any errors
            {
            }
        }

        void DoReceive()
        {	
	        // Check for Server Advertisement messages
            byte[] bArray = new byte[ServerAdvertisement_Message_Size];
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;
            try
            {
                iReceiveByteCount = m_LocalSocket_ForReceivingServerAdvertisements.ReceiveFrom(bArray, ref From_EndPoint);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Additional diagnostics can be placed here if necessary
                }
                return;
            }

            if (0 < iReceiveByteCount)
            {
                ServerAdvertisement_Message_PDU Message = DeSerialize_ServerAdvertisementMessage(bArray);

 			    switch(Message.eMessageType)
			    {
				    case MessageType.SERVER_ADVERTISEMENT:
                        if (false == m_bConnected_AppServer) // Only process the Server advertisement if not already connected to a server
                        {
                            GetServerAddress_FromServerAdvertisementMessage(Message);
                            ConnectButton.Enabled = true;
                            WriteStatusLine("Received SERVER_ADVERTISEMENT message");
                        }
					    break;
			    }
		    }
	
		    // Check for application messages
            byte[] bAppMessArray = new byte[Application_Message_Size];
            if (true == m_bConnected_AppServer)
            {
                try
                {
                    iReceiveByteCount = m_ConnectSocket.Receive(bAppMessArray);
                }
                catch (SocketException ex)
                {
                    if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                    {
                        // IGNORE - The exception was 'WOULD BLOCK' 
                    }
                    else
                    {
                        if (false == Is_TCP_SocketConnected(m_ConnectSocket))
                        {   // Connection to server lost
                            WriteStatusLine("Connection with server lost");
                            CloseSocketAndExit();
                        }
                    }
                    return;
                }

                if (0 < iReceiveByteCount)
                {
                    Application_Message_PDU AppMessage = DeSerialize_ApplicationMessage(bAppMessArray);

                    switch (AppMessage.eMessageType)
                    {
                        case MessageType.SEND_RECORD:
                            if (true == AppMessage.bCustNameValid) // Guard needed because each server returns only a partial row of data at the application level
                            {
                                string sStr = "";
                                ByteArrayTo_ASCII_String(AppMessage.cCustName, ref sStr);
                                CustomerName_textBox.Text = sStr;
                                SendUpdatedData_button.Enabled = true;
                                WriteStatusLine("Received SEND_RECORD message");
                            }
                            break;
                    }
                }
            }
        }

        ServerAdvertisement_Message_PDU DeSerialize_ServerAdvertisementMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a ServerAdvertisement_Message_PDU structure
            ServerAdvertisement_Message_PDU result_Message = new ServerAdvertisement_Message_PDU();
            //int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0 /*FieldOffset*/);
            result_Message.iSender_Random_ID = ByteArrayToInt32(bRecdByteArray, 4 /*FieldOffset*/);
            result_Message.iAddressFamily = bRecdByteArray[8]; // Value 2 = AF_INET
            // Deserialisation of port number
            // The message format and deserialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            result_Message.iPort = (256 * bRecdByteArray[10])/*1st of the 4 bytes*/ + bRecdByteArray[11]/*2nd of the 4 bytes*/;
            result_Message.iIP_Address_Byte_1 = bRecdByteArray[12];
            result_Message.iIP_Address_Byte_2 = bRecdByteArray[13];
            result_Message.iIP_Address_Byte_3 = bRecdByteArray[14];
            result_Message.iIP_Address_Byte_4 = bRecdByteArray[15];
            return result_Message;
        }

        Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
            Application_Message_PDU result_Message = new Application_Message_PDU();
            //int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0 /*FieldOffset*/);
            result_Message.iSender_Random_ID = ByteArrayToInt32(bRecdByteArray, 4 /*FieldOffset*/);
            result_Message.iCust_ID = ByteArrayToInt32(bRecdByteArray, 8 /*FieldOffset*/);
            result_Message.bCustNameValid = true;
            if (0 == bRecdByteArray[12])
            {
                result_Message.bCustNameValid = false;
            }
            result_Message.cCustName = new byte[MAX_CUST_NAME_LEN];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 13 /*source offset*/, result_Message.cCustName/*dest*/, 0/*cCustName offset*/, MAX_CUST_NAME_LEN/*count*/);
            return result_Message;
        }

        private Int32 GetCustIDFromControl()
        {
            Int32 iCustID = 0;
            if ("" == CustomerID_textBox.Text)
            {
                MessageBox.Show("Customer ID must be between 101 and 103", "Invalid Customer ID");
                return -1;
            }
            try
            {
                iCustID = Convert.ToInt32(CustomerID_textBox.Text);
                if (iCustID < 101 || iCustID > 103)
                {
                    MessageBox.Show("Customer ID must be between 101 and 103", "Invalid Customer ID");
                    return -1;
                }
            }
            catch
            {
                MessageBox.Show("Please enter a valid customer ID", "Invalid Customer ID");
                return -1;
            }
            return iCustID;
        }    
         
        private void RequestData_button_Click(object sender, EventArgs e)
        {
            Int32 iCustID = GetCustIDFromControl();
            if (-1 == iCustID)
            {
                return;
            }
            CustomerName_textBox.Text = "";   // Initially clear the data control
            Application_Message_PDU Message = new Application_Message_PDU();
            Message.eMessageType = MessageType.REQUEST_RECORD;
            Message.iCust_ID = iCustID;
            Message.bCustNameValid = false;
            if (true == m_bConnected_AppServer)
            {
                SendApplicationMessage(Message);  // Send REQUEST_RECORD message to AppServer
                WriteStatusLine("Sent REQUEST_RECORD message");
            }
        }

        private byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
        {   // Convert an Application_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[Application_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 0/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCust_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 8/*bByteArray offset*/, 4/*count*/);
            bByteArray[12] = 0; // false (bCustNameValid)
            if(true == Message.bCustNameValid)
            {
                bByteArray[12] = 1; // true (bCustNameValid)
            }
            if (null != Message.cCustName)
            {
                Buffer.BlockCopy(Message.cCustName /*src*/, 0/*source offset*/, bByteArray/*dest*/, 13/*cCustName offset*/, Message.cCustName.Length/*count*/);
            }
            return bByteArray;
        }

        private void SendUpdatedData_button_Click(object sender, EventArgs e)
        {
            Int32 iCustID = GetCustIDFromControl();
            if (-1 == iCustID)
            {
                return;
            }
            Application_Message_PDU Message = new Application_Message_PDU();
            Message.eMessageType = MessageType.DATA_UPDATE_FROM_CLIENT;
            Message.iCust_ID = iCustID;
            string sStr = TruncateString(CustomerName_textBox.Text, MAX_CUST_NAME_LEN);
            Message.cCustName = StringTo_ASCII_ByteArray(sStr);
            Message.bCustNameValid = true;
            if (true == m_bConnected_AppServer)
            {
                SendApplicationMessage(Message);  // Send REQUEST_RECORD message to AppServer
            }
            WriteStatusLine("Sent DATA_UPDATE_FROM_CLIENT message");
            SendUpdatedData_button.Enabled = false; // Data must be edited before the button can be pressed
        }

        void SendApplicationMessage(Application_Message_PDU Message)
        {
            byte[] bBuffer = Serialize_Application_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_ConnectSocket.Send(bBuffer);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    if (false == Is_TCP_SocketConnected(m_ConnectSocket))
                    {   // Connection to server lost
                        WriteStatusLine("Connection with server lost");
                        CloseSocketAndExit();
                    }
                }
            }
         }

        void WriteStatusLine(string sLine)
        {
            EventLog_listBox.Items.Add(sLine);
        }

        static bool Is_TCP_SocketConnected(Socket sSock)
        {
            if (sSock.ProtocolType != ProtocolType.Tcp)
            {   // This test is necessary because the Connected property may return true for a UDP
                // socket (which is meaningless as it is alsways 'connectionless')
                return false;
            }
            // Note, the Connected property alone is not reliable. It indicates the connected state of the socket
            // at the time of executing the most recent Send or Receive operation
            return !((sSock.Poll(1000, SelectMode.SelectRead) && (sSock.Available == 0)) || !sSock.Connected);
        }

        private void Done_Button_Click(object sender, EventArgs e)
        {
            CloseSocketAndExit();
        }

        void CloseSocketAndExit()
        {
            CloseConnectionAndSocket();
            Application.Exit();
        }

        void CloseConnectionAndSocket()
        {
            StopTimer();
            try
            {
                m_ConnectSocket.Shutdown(SocketShutdown.Both);	// Close the connection (in both directions)
            }
            catch
            {   // Generic approach to dealing with situations such as socket not connected
            }
            try
            {
                m_ConnectSocket.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            //Encoding.ASCII.GetBytes(sStr);
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if (sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }
    }
}
