﻿/*	File			Replication_Server.cs
	Purpose			Replication demonstration
	Author			Richard John Anthony
	Date			February 2015
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace ReplicationServer
{
    public partial class Replication_Server : Form
    {
        public const Int16 REPLICATION_EXTERNAL_PORT = 8006;  // Decimal (Client to Server)
        public const Int16 REPLICATION_UPDATE_PORT = 8007;	// Decimal (Server to Server - Update propagation)
        public const Int16 REPLICATION_HEARTBEAT_PORT = 8008; // Decimal (Server to Server - Master heartbeat broadcasts)
        public const Int16 REPLICATION_SERVER_ADVERTISEMENT_PORT = 8009; // Decimal (Server to client)

        public const Int16 MAX_CUST_NAME_LEN = 20;

        public enum ServerState {MASTER, BACKUP};
	        // Default state for server instances is BACKUP, in this state:
	        //                  - listen for Master heartbeats, and update propagations
	        //                  - do not accept client connections
	        //                  - elevate to MASTER state if three consecutive master heartbeats missed

	        // MASTER state, in this state:
	        //                  - send periodic Master heartbeats (broadcast)
	        //                  - listen for Master heartbeats
	        //                  - accept client connections
	        //                  - demote to BACKUP state if master heartbeat from another server instance is detected

        public enum MessageType {SERVER_ADVERTISEMENT, HEARTBEAT, REQUEST_RECORD, SEND_RECORD, 
            DATA_UPDATE_FROM_CLIENT, UPDATE_PROPAGATION_FROM_MASTER, CLIENT_DISCONNECT};

        // Message structures (these will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Application_Message_PDU 
        {
	        public MessageType eMessageType;
	        public int iSender_Random_ID;				// Needed so that UPDATE_PROPAGATION_FROM_MASTER messages from self can be detected and ignored
	        public int iCust_ID;
	        public bool bCustNameValid;                 // Single byte but will need to pad to 4-byte boundary (i.e. add 3 bytes padding)
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_CUST_NAME_LEN + 1))]
	        public byte[] cCustName;	
        };

        public const Int16 Application_Message_Size = 36;

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Heartbeat_Message_PDU
        {
	        public MessageType eMessageType;
	        public int iSender_Random_ID;				// Needed so that messages from self can be detected and ignored
	        public int iSender_Random_ID_Incremented;	// Used to differentiate genuine heartbeat messages from any other messages that just happen to arrive on this port
        };

        public const Int16 Heartbeat_Message_Size = 12;

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct ServerAdvertisement_Message_PDU
        {
	        public MessageType eMessageType;
	        public Int32 iSender_Random_ID;				
            public Int32 iAddressFamily;
            public Int32 iPort; // Port values are actually Int16, but encoded as Int32 in this message format for compatibility with the pre-existing C++ version
            public byte iIP_Address_Byte_1; 
            public byte iIP_Address_Byte_2; 
            public byte iIP_Address_Byte_3; 
            public byte iIP_Address_Byte_4; 
        };

        public const Int16 ServerAdvertisement_Message_Size = 24;  // For compatibility with C++ version

	    Socket m_ListenSock;  // Listen for client connections
	    Socket m_ConnectSock; // Connection to Client
	    Socket m_HeartbeatBroadcastSock; // Send heartbeat broadcasts
	    Socket m_HeartbeatReceiveSock; // Recieve heartbeat broadcasts
	    Socket m_ServerAdvertisementBroadcastSock;
	    Socket m_BroadcastPropagatedUpdateSock; // Send heartbeat broadcasts
	    Socket m_ReceivePropagatedUpdateSock; // Recieve propagated database update broadcasts

 	    IPEndPoint m_Local_IPEndPoint;
	    IPEndPoint m_BROADCAST_Heartbeat_IPEndPoint;
	    IPEndPoint m_ReceiveHeartbeat_IPEndPoint;
	    IPEndPoint m_BROADCAST_ServerAdvertisement_IPEndPoint;
	    IPEndPoint m_BROADCAST_PropagatedUpdate_IPEndPoint;	
	    IPEndPoint m_ReceivePropagatedUpdates_IPEndPoint;


	    ServerState m_ServerState;
	    int m_iHeartbeatMissedCount;
	    Int32 m_iServer_Random_ID;
	    bool m_bConnected;

        System.Windows.Forms.Timer m_Timer_Application;
        System.Windows.Forms.Timer m_Timer_Send_Heartbeat;
        System.Windows.Forms.Timer m_Timer_Receive_Heartbeat;
        System.Windows.Forms.Timer m_Timer_Send_ServerAdvertisement;
	    bool m_bHeartbeatTimerRunning;
	    bool m_bServerAdvertisementTimerRunning;

        // Data Row and Table structures (hard coded data serves as mock of real database)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct DataRow_struct 
        {
	        public int iCust_ID;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_CUST_NAME_LEN + 1))]
	        public byte[] cCustName;	
        };

        // DataTable_struct
	    DataRow_struct[] m_DataRowArray; // Array of data row structures

        public Replication_Server()
        {
            InitializeComponent();
            m_DataRowArray = new DataRow_struct[3];
            GetLocalHostAddressDetails();
 
            Random rnd = new Random();
 	        m_iServer_Random_ID = rnd.Next(1,10000);
            ServerID_textBox.Text = m_iServer_Random_ID.ToString();
            
            m_bHeartbeatTimerRunning = false;
	        m_bServerAdvertisementTimerRunning = false;	
	        m_bConnected = false;

	        InitialiseDatabase();
	        DisplayDataTable();
	        EnableServer();
	        Initialise_Send_Heartbeat_Socket();
	        Initialise_Send_ServerAdvertisement_Socket();
	        Initialise_Broadcast_Update_propagation_Socket();
	        Enable_ReceiveHeartbeatMessages();
            Start_Timer_Receive_Heartbeat();	// Initiate listening for heartbeat messages	
	        Enable_ReceivePropagatedUpdateMessages();

	        Transition_To_BACKUP_State(true);
        }

        void InitialiseDatabase()	// Populate DataTable (hardcoded data)
        {
            m_DataRowArray[0].iCust_ID = 101;
            m_DataRowArray[0].cCustName = StringTo_ASCII_ByteArray("Fred");
            m_DataRowArray[1].iCust_ID = 102;
            m_DataRowArray[1].cCustName = StringTo_ASCII_ByteArray("Jane");
            m_DataRowArray[2].iCust_ID = 103;
            m_DataRowArray[2].cCustName = StringTo_ASCII_ByteArray("Rita");
        }

        void UpdateDatabase(int iCustID_AsIndex, string sCustName)	// Upate DataTable
        {
	        for(int iIndex = 0; iIndex < 3; iIndex++)
	        {
		        if(iCustID_AsIndex == m_DataRowArray[iIndex].iCust_ID)
		        {
                    string sStr = TruncateString(sCustName, MAX_CUST_NAME_LEN);
                    m_DataRowArray[iIndex].cCustName = StringTo_ASCII_ByteArray(sStr);
                    DisplayDataTable();
			        return;
		        }
	        }		
        }
 
        void DisplayDataTable()
        {	
	        CustID_listBox.Items.Clear();
	        CustName_listBox.Items.Clear();

	        for(int iIndex = 0; iIndex < 3; iIndex ++)
	        {
		        CustID_listBox.Items.Add(m_DataRowArray[iIndex].iCust_ID.ToString());
                string sStr = "";
                ByteArrayTo_ASCII_String(m_DataRowArray[iIndex].cCustName, ref sStr);
                CustName_listBox.Items.Add(sStr);
            }
        }

        void GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {   // Create the local IPEndPoint using the first IPv4 address and the pre-defined port
                    m_Local_IPEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], REPLICATION_EXTERNAL_PORT);
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    string sStr = string.Format("Server Address (recv application messages): {0}.{1}.{2}.{3}  Port:{4}", 
                        bAddressBytes[0].ToString(), bAddressBytes[1].ToString(), bAddressBytes[2].ToString(),
                        bAddressBytes[3].ToString(), REPLICATION_EXTERNAL_PORT);
                    WriteStatusLine(sStr);

                    m_ReceiveHeartbeat_IPEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], REPLICATION_HEARTBEAT_PORT);
                    bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    sStr = string.Format("Server Address (recv heartbeat): {0}.{1}.{2}.{3}  Port:{4}", 
                        bAddressBytes[0].ToString(), bAddressBytes[1].ToString(), bAddressBytes[2].ToString(),
                        bAddressBytes[3].ToString(), REPLICATION_HEARTBEAT_PORT);
                    WriteStatusLine(sStr);
                    
                    m_ReceivePropagatedUpdates_IPEndPoint = new IPEndPoint(ipHostAddress[iAddrIndex], REPLICATION_UPDATE_PORT);
                    bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    sStr = string.Format("Server Address (recv propagated updates): {0}.{1}.{2}.{3}  Port:{4}", 
                        bAddressBytes[0].ToString(), bAddressBytes[1].ToString(), bAddressBytes[2].ToString(),
                        bAddressBytes[3].ToString(), REPLICATION_UPDATE_PORT);
                    WriteStatusLine(sStr);
                   
                    return;
                }
            }
	        MessageBox.Show("Failed to get local address!","Replication Server");
            CloseSocketAndExit();
        }
        
        void InitialiseAddressStructAndPort_For_BROADCAST_Heartbeat()
        {	// Initialise the IP address
	        m_BROADCAST_Heartbeat_IPEndPoint = new IPEndPoint(IPAddress.Broadcast, REPLICATION_HEARTBEAT_PORT);
        }

        void InitialiseAddressStructAndPort_For_BROADCAST_ServerAdvertisement()
        {	// Initialise the IP address
	        m_BROADCAST_ServerAdvertisement_IPEndPoint = 
                                new IPEndPoint(IPAddress.Broadcast, REPLICATION_SERVER_ADVERTISEMENT_PORT);
        }

        void InitialiseAddressStructAndPort_For_BROADCAST_UpdatePropagationMessages()
        {	// Initialise the IP address
	        m_BROADCAST_PropagatedUpdate_IPEndPoint = 
                                new IPEndPoint(IPAddress.Broadcast, REPLICATION_UPDATE_PORT);
        }

        void Accept()
        {
            try
            {
                m_ConnectSock = m_ListenSock.Accept();
                // Set the newly-created connection socket to Non-Blocking
                m_ConnectSock.Blocking = false;
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    MessageBox.Show("Accept failed","Replication Server");
                }
                return;
            }
            m_bConnected = true;
            string sStr = string.Format("Client {0} Connected", m_ConnectSock.RemoteEndPoint.ToString());
            WriteStatusLine(sStr);
        }

        void EnableServer()  // Invoked on startup
        {
            try
            {   // Create Listen Socket and set it NonBlocking
                m_ListenSock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ListenSock.Blocking = false;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Replication Server");
                CloseSocketAndExit();
            }

            try
            {
                m_ListenSock.Bind(m_Local_IPEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Bind failed !\nPort is probably in use","Replication Server");
                CloseSocketAndExit();
            }

	        // Listen - for TCP connection requests from clients
            try
            {
                m_ListenSock.Listen(5);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Listen failed","Replication Server");
                CloseSocketAndExit();
            }
            Start_Timer_Application();   // As all activity is non-blocking
        }

        void Enable_ReceiveHeartbeatMessages()  // Invoked on startup
        {
            try
            {   // Create Socket and set it NonBlocking
                m_HeartbeatReceiveSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_HeartbeatReceiveSock.Blocking = false;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Replication Server");
                CloseSocketAndExit();
            }
    
            try
            {
                m_HeartbeatReceiveSock.Bind(m_ReceiveHeartbeat_IPEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Bind failed !\nPort is probably in use","Replication Server");
                CloseSocketAndExit();
            }
            WriteStatusLine("Bind succeeded (Heartbeat port)");
	    }

        void Enable_ReceivePropagatedUpdateMessages()  // Invoked on startup
        {
            try
            {   // Create Socket and set it NonBlocking
                m_ReceivePropagatedUpdateSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_ReceivePropagatedUpdateSock.Blocking = false;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Replication Server");
                CloseSocketAndExit();
            }
    
            try
            {
                m_ReceivePropagatedUpdateSock.Bind(m_ReceivePropagatedUpdates_IPEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Bind failed !\nPort is probably in use","Replication Server");
                CloseSocketAndExit();
            }
            WriteStatusLine("Bind succeeded (Propagated updates port)");
        }

        void Transition_To_MASTER_State()   // replace all other transitions !
        {
            Start_Timer_Send_ServerAdvertisement();	// Start timer to govern sending of Server Advertisement messages
            Start_Timer_Send_Heartbeat();			// Start timer to govern sending of Heartbeat messages
	        m_ServerState = ServerState.MASTER;
	        Update_ServerState_Display();
	        m_iHeartbeatMissedCount = 0;
	        HeartbeatMissedCount_textBox.Text = m_iHeartbeatMissedCount.ToString();

	        WriteStatusLine("Promoted to MASTER State (no other MASTER detected)");
        }

        void Transition_To_BACKUP_State(bool bSilent /*dont display 'demoted' message when called during initialisation*/)
        {
	        if(true == m_bHeartbeatTimerRunning)
	        {
		        m_Timer_Send_Heartbeat.Enabled = false;				// Stop sending periodic heartbeats
		        m_bHeartbeatTimerRunning = false;
	        }

	        if(true == m_bServerAdvertisementTimerRunning)
	        {
		        m_Timer_Send_ServerAdvertisement.Enabled = false;	// Stop sending Server Advertisements
		        m_bServerAdvertisementTimerRunning = false;
	        }

	        m_ServerState = ServerState.BACKUP;	// Default state is backup - in this state listen for Master heartbeats 
	        Update_ServerState_Display();
	        m_iHeartbeatMissedCount = 0;
	        HeartbeatMissedCount_textBox.Text = m_iHeartbeatMissedCount.ToString();

	        if(false == bSilent)
	        {
		        WriteStatusLine("Demoted to BACKUP status (other MASTER detected)");
	        }
        }

       void Start_Timer_Application()
        {
            m_Timer_Application = new System.Windows.Forms.Timer();
            m_Timer_Application.Interval = 100; // milliseconds
            m_Timer_Application.Tick += OnTimedEvent_Timer_Application;
            m_Timer_Application.Enabled = true;
        }

        void OnTimedEvent_Timer_Application(Object myObject, EventArgs myEventArgs)
        {
			DoReceive_Heartbeat(); // Check for receipt of Heartbeat messages
			DoReceive();
			if(false == m_bConnected)
			{
				Accept();
			}
        }

       void Start_Timer_Send_Heartbeat()
        {
            m_Timer_Send_Heartbeat = new System.Windows.Forms.Timer();
            m_Timer_Send_Heartbeat.Interval = 3000; // milliseconds
            m_Timer_Send_Heartbeat.Tick += OnTimedEvent_Timer_Send_Heartbeat;
            m_Timer_Send_Heartbeat.Enabled = true;
	        m_bHeartbeatTimerRunning = true;
        }

        void OnTimedEvent_Timer_Send_Heartbeat(Object myObject, EventArgs myEventArgs)
        {
			if(ServerState.BACKUP == m_ServerState)
			{
				// No action, BACKUP server instances do not send Heartbeats
			}
			if(ServerState.MASTER == m_ServerState)
			{
				WriteStatusLine("Send heartbeat");
				Send_Heartbeat();       // Broadcast heartbeat message		
			}
        }

       void Start_Timer_Receive_Heartbeat()
        {
            m_Timer_Receive_Heartbeat = new System.Windows.Forms.Timer();
            m_Timer_Receive_Heartbeat.Interval = 3300; // 3.3 second interval allowed between receiving heartbeat messages (avoid a race if same time interval is used as sender)
            m_Timer_Receive_Heartbeat.Tick += OnTimedEvent_Timer_Receive_Heartbeat;
            m_Timer_Receive_Heartbeat.Enabled = true;
        }

        void OnTimedEvent_Timer_Receive_Heartbeat(Object myObject, EventArgs myEventArgs)
        {
			if(ServerState.BACKUP == m_ServerState)
			{
				m_iHeartbeatMissedCount ++;
				HeartbeatMissedCount_textBox.Text = m_iHeartbeatMissedCount.ToString();
				string sStr = string.Format("Heartbeat missed! count = {0}", m_iHeartbeatMissedCount);
				WriteStatusLine(sStr);
				if(3 == m_iHeartbeatMissedCount)
				{
					Transition_To_MASTER_State();
				}
			}
			if(ServerState.MASTER == m_ServerState)
			{
				// No action as Heartbeat was not expected, so its absence requires no action
			}
        }

        void Start_Timer_Send_ServerAdvertisement()
        {
            m_Timer_Send_ServerAdvertisement = new System.Windows.Forms.Timer();
            m_Timer_Send_ServerAdvertisement.Interval = 5000;	// 5 second interval between sending ServerAdvertisement messages
            m_Timer_Send_ServerAdvertisement.Tick += OnTimedEvent_Timer_Send_ServerAdvertisement;
            m_Timer_Send_ServerAdvertisement.Enabled = true;
	        m_bServerAdvertisementTimerRunning = true;
        }

        void OnTimedEvent_Timer_Send_ServerAdvertisement(Object myObject, EventArgs myEventArgs)
        {
			if(ServerState.BACKUP == m_ServerState)
			{
				// No action, BACKUP server instances do not send Server Advertisements
			}
			if(ServerState.MASTER == m_ServerState)
			{
				WriteStatusLine("Send Server Advertisement");
				Send_ServerAdvertisement(); // Broadcast Server Advertisement message					
			}
        }

        void Initialise_Send_ServerAdvertisement_Socket()
        {
            try
            {
                m_ServerAdvertisementBroadcastSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_ServerAdvertisementBroadcastSock.EnableBroadcast = true;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Replication Client");
                CloseSocketAndExit();
            }
	        InitialiseAddressStructAndPort_For_BROADCAST_ServerAdvertisement();
        }

        void Initialise_Send_Heartbeat_Socket()
        {
            try
            {
                m_HeartbeatBroadcastSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_HeartbeatBroadcastSock.EnableBroadcast = true;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Replication Client");
                CloseSocketAndExit();
            }
	        InitialiseAddressStructAndPort_For_BROADCAST_Heartbeat();
        }

        void Initialise_Broadcast_Update_propagation_Socket()
        {
            try
            {
                m_BroadcastPropagatedUpdateSock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_BroadcastPropagatedUpdateSock.EnableBroadcast = true;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "Replication Client");
                CloseSocketAndExit();
            }
	        InitialiseAddressStructAndPort_For_BROADCAST_UpdatePropagationMessages();
        }

        void Send_Heartbeat()
        {	
	        Heartbeat_Message_PDU Heartbeat_Message = new Heartbeat_Message_PDU();
	        Heartbeat_Message.eMessageType = MessageType.HEARTBEAT;
	        Heartbeat_Message.iSender_Random_ID = m_iServer_Random_ID;
	        Heartbeat_Message.iSender_Random_ID_Incremented = m_iServer_Random_ID + 1;

            byte[] bBuffer = Serialize_Heartbeat_Message_PDU(Heartbeat_Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_HeartbeatBroadcastSock.SendTo(bBuffer, m_BROADCAST_Heartbeat_IPEndPoint);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Place diagnostics here if necessay
                }
            }
        }

        void Send_ServerAdvertisement()
        {	
	        ServerAdvertisement_Message_PDU ServerAdvertisement_Message = new ServerAdvertisement_Message_PDU();
	        ServerAdvertisement_Message.eMessageType = MessageType.SERVER_ADVERTISEMENT;
	        ServerAdvertisement_Message.iSender_Random_ID = m_iServer_Random_ID;

            byte[] bAddress = m_Local_IPEndPoint.Address.GetAddressBytes();
            ServerAdvertisement_Message.iIP_Address_Byte_1 = bAddress[0];
            ServerAdvertisement_Message.iIP_Address_Byte_2 = bAddress[1];
            ServerAdvertisement_Message.iIP_Address_Byte_3 = bAddress[2];
            ServerAdvertisement_Message.iIP_Address_Byte_4 = bAddress[3];
            ServerAdvertisement_Message.iAddressFamily = 2; // Equivallent to C++ AF_INET     
            ServerAdvertisement_Message.iPort = REPLICATION_EXTERNAL_PORT;    

            byte[] bBuffer = Serialize_ServerAdvertisement_Message_PDU(ServerAdvertisement_Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_ServerAdvertisementBroadcastSock.SendTo(bBuffer, m_BROADCAST_ServerAdvertisement_IPEndPoint);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Place diagnostics here if necessay
                }
            }
        }

        void SendApplicationMessage(Application_Message_PDU Message)
        {
            byte[] bBuffer = Serialize_Application_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_ConnectSock.Send(bBuffer);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    if (false == Is_TCP_SocketConnected(m_ConnectSock))
                    {   // Connection to server lost
                        WriteStatusLine("Connection lost");
                        CloseSocketAndExit();
                    }
                }
            }
         }

        void Send_DataRow_Message(int iCustID)
        {
	        Application_Message_PDU Message = new Application_Message_PDU();
	        Message.eMessageType = MessageType.SEND_RECORD;
	        Message.iCust_ID = iCustID;
	        Message.cCustName = StringTo_ASCII_ByteArray("------No Match------");

	        for(int iIndex = 0; iIndex < 3; iIndex++)
	        {
		        if(iCustID == m_DataRowArray[iIndex].iCust_ID)
		        {
                    Buffer.BlockCopy(m_DataRowArray[iIndex].cCustName/*src*/, 0 /*offset*/, Message.cCustName/*dest*/, 0/*offset*/, m_DataRowArray[iIndex].cCustName.Length/*count*/);
                }
	        }

	        Message.bCustNameValid = true;
	        SendApplicationMessage(Message);
            string sCust = "";
            ByteArrayTo_ASCII_String(Message.cCustName, ref sCust);
            string sStr = string.Format("Data Record {0} {1} sent", Message.iCust_ID, sCust);
	        WriteStatusLine(sStr);
        }

        void DoReceive()
        {		
            // Check for Database update propagation messages (a Application_Message_PDU from MASTER instance of service)
            byte[] bArray = new byte[Application_Message_Size];
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;
            try
            {
                iReceiveByteCount = m_ReceivePropagatedUpdateSock.ReceiveFrom(bArray, ref From_EndPoint);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Additional diagnostics can be placed here if necessary
                }
                iReceiveByteCount = 0; // Force skip of next section, to check for messages on the m_ConnectSock socket
            }

            if (0 < iReceiveByteCount)
            {
                Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
			    switch(Message.eMessageType)
			    {
				    case MessageType.UPDATE_PROPAGATION_FROM_MASTER:
					    if(m_iServer_Random_ID != Message.iSender_Random_ID)
					    { // Ignore echos from self
                            string sCust = "";
                            ByteArrayTo_ASCII_String(Message.cCustName, ref sCust);
                            string sStr = string.Format("Propagated data update recd from MASTER server: {0}, {1}", Message.iCust_ID,sCust);
						    WriteStatusLine(sStr);
						    UpdateDatabase(Message.iCust_ID, sCust);
					    }
					    break;
			    }
		    }

	        // Check for Application messages (from Client)
	        if(true == m_bConnected)
	        {
                iReceiveByteCount = 0;
                try
                {
                    iReceiveByteCount = m_ConnectSock.Receive(bArray);
                }
                catch (SocketException ex)
                {
                    if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                    {
                        // IGNORE - The exception was 'WOULD BLOCK' 
                    }
                    else
                    {
                        // Additional diagnostics can be placed here if necessary
                    }
                    return;
                }
			    if(0 < iReceiveByteCount)
			    {
                    Application_Message_PDU Message = DeSerialize_ApplicationMessage(bArray);
			        string sStr;
			        switch(Message.eMessageType)
			        {
				       case MessageType.REQUEST_RECORD:
					        sStr = string.Format("Data record {0} requested", Message.iCust_ID);
					        WriteStatusLine(sStr);
					        Send_DataRow_Message(Message.iCust_ID);
					        break;		

				        case MessageType.DATA_UPDATE_FROM_CLIENT:
                            string sCust = "";
                            ByteArrayTo_ASCII_String(Message.cCustName, ref sCust);
					        sStr = string.Format("Client sent data update {0}, {1}",Message.iCust_ID, sCust);
					        WriteStatusLine(sStr);
					        UpdateDatabase(Message.iCust_ID, sCust);
					        PropagateDatabaseUpdate_To_BACKUP_ServerInstance(Message.iCust_ID, Message.cCustName);   // Forward the database update to the BACKUP server
					        break;

				        case MessageType.CLIENT_DISCONNECT:
					        m_bConnected = false;
					        CloseConnection();
					        WriteStatusLine("Client disconnected");
					        break;
			        }
		        }
	        }
        }

        void PropagateDatabaseUpdate_To_BACKUP_ServerInstance(int iCustID, byte[] cCustName)   // Forward the database update to the BACKUP server
        {
	        Application_Message_PDU Message = new Application_Message_PDU();
	        Message.eMessageType = MessageType.UPDATE_PROPAGATION_FROM_MASTER;
	        Message.iSender_Random_ID = m_iServer_Random_ID;
	        Message.iCust_ID = iCustID;
            Message.cCustName = new byte[MAX_CUST_NAME_LEN + 1];
            Buffer.BlockCopy(cCustName/*src*/, 0/*offset*/, Message.cCustName/*dest*/, 0/*offset*/, cCustName.Length/*count*/);
            Message.bCustNameValid = true;

            byte[] bBuffer = Serialize_Application_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_BroadcastPropagatedUpdateSock.SendTo(bBuffer, m_BROADCAST_PropagatedUpdate_IPEndPoint);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessay
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Place diagnostics here if necessay
                }
            }

            string sCust = "";
            ByteArrayTo_ASCII_String(Message.cCustName, ref sCust);
            string sStr = string.Format("Data propagation update sent: {0} {1}", Message.iCust_ID, sCust);
	        WriteStatusLine(sStr);
        }

        void DoReceive_Heartbeat()
        {	// Check for heartbeat messages (UDP broadcast from a MASTER server instance)
            byte[] bArray = new byte[Heartbeat_Message_Size];
            int iReceiveByteCount = 0;
            IPEndPoint From_IPEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
            EndPoint From_EndPoint = (EndPoint)From_IPEndPoint;
            try
            {
                iReceiveByteCount = m_HeartbeatReceiveSock.ReceiveFrom(bArray, ref From_EndPoint);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {
                    // Additional diagnostics can be placed here if necessary
                }
                return;
            }

            if (0 < iReceiveByteCount)
            {
                Heartbeat_Message_PDU Message = DeSerialize_HeartbeatMessage(bArray);
			    switch(Message.eMessageType)
			    {
				    case MessageType.HEARTBEAT:
					    if((Message.iSender_Random_ID + 1) == Message.iSender_Random_ID_Incremented)
					    { // Message is a genuine heartbeat message, not spurious message on heartbeat port (important check because broadcast is used, not a connection)
						    if(m_iServer_Random_ID != Message.iSender_Random_ID)
						    { // Ignore echos from self
							    string sStr = string.Format("Heartbeat message received (from server ID:{0})",Message.iSender_Random_ID);
							    WriteStatusLine(sStr);
							    m_iHeartbeatMissedCount = 0;
							    HeartbeatMissedCount_textBox.Text = m_iHeartbeatMissedCount.ToString();
							    if(ServerState.MASTER == m_ServerState)
							    {
								    Transition_To_BACKUP_State(false);
							    }
							    else
							    {
                                    m_Timer_Receive_Heartbeat.Stop();
                                    m_Timer_Receive_Heartbeat.Start(); // Restart heartbeat timer so that it does not timeout
							    }
						    }
					    }
				    break;
			    }
		    }
	    }

        private void ClearLog_button_Click(object sender, EventArgs e)
        {
            EventLog_listBox.Items.Clear();
        }

        void Update_ServerState_Display()
        {
	        ServerState_textBox.Text = m_ServerState.ToString();
	        HeartbeatMissedCount_textBox.Text = m_iHeartbeatMissedCount.ToString();
        }

        ServerAdvertisement_Message_PDU DeSerialize_ServerAdvertisementMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a ServerAdvertisement_Message_PDU structure
            ServerAdvertisement_Message_PDU result_Message = new ServerAdvertisement_Message_PDU();
            //int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0 /*FieldOffset*/);
            result_Message.iSender_Random_ID = ByteArrayToInt32(bRecdByteArray, 4 /*FieldOffset*/);
            result_Message.iAddressFamily = bRecdByteArray[8]; // Value 2 = AF_INET
            // Deserialisation of port number
            // The message format and deserialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            result_Message.iPort = (256 * bRecdByteArray[10])/*1st of the 4 bytes*/ + bRecdByteArray[11]/*2nd of the 4 bytes*/;
            result_Message.iIP_Address_Byte_1 = bRecdByteArray[12];
            result_Message.iIP_Address_Byte_2 = bRecdByteArray[13];
            result_Message.iIP_Address_Byte_3 = bRecdByteArray[14];
            result_Message.iIP_Address_Byte_4 = bRecdByteArray[15];
            return result_Message;
        }

        Heartbeat_Message_PDU DeSerialize_HeartbeatMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a Heartbeat_Message_PDU structure
            Heartbeat_Message_PDU result_Message = new Heartbeat_Message_PDU();
            //int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0/*FieldOffset*/);
            result_Message.iSender_Random_ID = ByteArrayToInt32(bRecdByteArray, 4/*FieldOffset*/);
            result_Message.iSender_Random_ID_Incremented = ByteArrayToInt32(bRecdByteArray, 8/*FieldOffset*/);
            return result_Message;
        }

        Application_Message_PDU DeSerialize_ApplicationMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to an Application_Message_PDU structure
            Application_Message_PDU result_Message = new Application_Message_PDU();
            //int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, 0 /*FieldOffset*/);
            result_Message.iSender_Random_ID = ByteArrayToInt32(bRecdByteArray, 4 /*FieldOffset*/);
            result_Message.iCust_ID = ByteArrayToInt32(bRecdByteArray, 8 /*FieldOffset*/);
            result_Message.bCustNameValid = true;
            if (0 == bRecdByteArray[12])
            {
                result_Message.bCustNameValid = false;
            }
            result_Message.cCustName = new byte[MAX_CUST_NAME_LEN+1];
            Buffer.BlockCopy(bRecdByteArray/*src*/, 13 /*source offset*/, result_Message.cCustName/*dest*/, 0/*cCustName offset*/, MAX_CUST_NAME_LEN/*count*/);
            return result_Message;
        }

        private byte[] Serialize_ServerAdvertisement_Message_PDU(ServerAdvertisement_Message_PDU Message)
        {   // Convert an ServerAdvertisement_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[ServerAdvertisement_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 0/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iSender_Random_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 4/*bByteArray offset*/, 4/*count*/);
            bByteArray[8] = (byte)Message.iAddressFamily;   // For C++ compatibility, value 2 = AF_INET

            // Serialisation of port number
            // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            bByteArray[10] = (byte)(Message.iPort / 256);
            bByteArray[11] = (byte)(Message.iPort % 256);
            bByteArray[12] = Message.iIP_Address_Byte_1;
            bByteArray[13] = Message.iIP_Address_Byte_2;
            bByteArray[14] = Message.iIP_Address_Byte_3;
            bByteArray[15] = Message.iIP_Address_Byte_4;
            return bByteArray;
        }

        private byte[] Serialize_Heartbeat_Message_PDU(Heartbeat_Message_PDU Message)
        {   // Convert an Heartbeat_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[Heartbeat_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 0/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iSender_Random_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 4/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iSender_Random_ID_Incremented);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 8/*bByteArray offset*/, 4/*count*/);
            return bByteArray;
        }

        private byte[] Serialize_Application_Message_PDU(Application_Message_PDU Message)
        {   // Convert an Application_Message_PDU structure into a byte array
            byte[] bByteArray = new byte[Application_Message_Size];        // Create a byte array to hold the serialised form of the message PDU
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 0/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iSender_Random_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 4/*bByteArray offset*/, 4/*count*/);
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCust_ID);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, 8/*bByteArray offset*/, 4/*count*/);
            bByteArray[12] = 0; // false (bCustNameValid)
            if(true == Message.bCustNameValid)
            {
                bByteArray[12] = 1; // true (bCustNameValid)
            }
            if (null != Message.cCustName)
            {
                Buffer.BlockCopy(Message.cCustName /*src*/, 0/*source offset*/, bByteArray/*dest*/, 13/*cCustName offset*/, Message.cCustName.Length/*count*/);
            }
            return bByteArray;
        }

        void WriteStatusLine(string sLine)
        {
            EventLog_listBox.Items.Add(sLine);
        }

        static bool Is_TCP_SocketConnected(Socket sSock)
        {
            if (sSock.ProtocolType != ProtocolType.Tcp)
            {   // This test is necessary because the Connected property may return true for a UDP
                // socket (which is meaningless as it is alsways 'connectionless')
                return false;
            }
            // Note, the Connected property alone is not reliable. It indicates the connected state of the socket
            // at the time of executing the most recent Send or Receive operation
            return !((sSock.Poll(1000, SelectMode.SelectRead) && (sSock.Available == 0)) || !sSock.Connected);
        }

        private void Done_Button_Click(object sender, EventArgs e)
        {
            CloseSocketAndExit();
        }

        void CloseSocketAndExit()
        {
            CloseConnectionAndSocket();
            Application.Exit();
        }

        void CloseConnection()
        {
            try
            {
                m_ConnectSock.Shutdown(SocketShutdown.Both);	// Close the connection (in both directions)
            }
            catch
            {   // Generic approach to dealing with situations such as socket not connected
            }
        }

        void CloseConnectionAndSocket()
        {
            m_Timer_Application.Enabled = false;
            CloseConnection();
            try
            {
                m_ConnectSock.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if (sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }
    }
}
